import pytest

# This test skips efa-direct because it requests FI_TAGGED
@pytest.mark.functional
def test_av_xfer(cmdline_args):
    from common import ClientServerTest
    test = ClientServerTest(cmdline_args, "fi_av_xfer -e rdm", fabric="efa")
    test.run()



from efa.efa_common import efa_run_client_server_test
from common import perf_progress_model_cli

import pytest
import copy


@pytest.mark.functional
def test_rdm_efa(cmdline_args, completion_semantic, fabric):
    from common import ClientServerTest
    test = ClientServerTest(cmdline_args, "fi_rdm", completion_semantic=completion_semantic, fabric=fabric)
    test.run()

# This test skips efa-direct because it requests FI_ORDER_SAS
@pytest.mark.functional
def test_rdm_bw_functional_efa(cmdline_args, completion_semantic):
    from common import ClientServerTest
    test = ClientServerTest(cmdline_args, "fi_flood -e rdm -v -T 1", completion_semantic=completion_semantic, fabric="efa")
    test.run()

@pytest.mark.parametrize("iteration_type",
                         [pytest.param("short", marks=pytest.mark.short),
                          pytest.param("standard", marks=pytest.mark.standard)])
def test_rdm_pingpong(cmdline_args, iteration_type, completion_semantic,
                      memory_type_bi_dir, completion_type, direct_message_size, fabric):
    command = "fi_rdm_pingpong"  + " " + perf_progress_model_cli
    efa_run_client_server_test(cmdline_args, command, iteration_type,
                               completion_semantic, memory_type_bi_dir,
                               direct_message_size if fabric == "efa-direct" else "all",
                               completion_type=completion_type, fabric=fabric)

@pytest.mark.functional
def test_rdm_pingpong_range(cmdline_args, completion_semantic, memory_type_bi_dir, message_size, direct_message_size, fabric):
    efa_run_client_server_test(cmdline_args, "fi_rdm_pingpong", "short",
                               completion_semantic, memory_type_bi_dir,
                               direct_message_size if fabric == "efa-direct" else message_size, fabric=fabric)

@pytest.mark.functional
def test_rdm_pingpong_no_inject_range(cmdline_args, completion_semantic, inject_message_size, direct_message_size, fabric):
    efa_run_client_server_test(cmdline_args, "fi_rdm_pingpong -j 0", "short",
                               completion_semantic, "host_to_host",
                               direct_message_size if fabric == "efa-direct" else inject_message_size, fabric=fabric)

# efa-direct does not support tagged
@pytest.mark.parametrize("iteration_type",
                         [pytest.param("short", marks=pytest.mark.short),
                          pytest.param("standard", marks=pytest.mark.standard)])
def test_rdm_tagged_pingpong(cmdline_args, iteration_type, completion_semantic, memory_type_bi_dir, completion_type):
    command = "fi_rdm_tagged_pingpong"  + " " + perf_progress_model_cli
    efa_run_client_server_test(cmdline_args, command, iteration_type,
                               completion_semantic, memory_type_bi_dir, "all", completion_type=completion_type,
                               fabric="efa")

@pytest.mark.functional
def test_rdm_tagged_pingpong_range(cmdline_args, completion_semantic, memory_type_bi_dir, message_size):
    efa_run_client_server_test(cmdline_args, "fi_rdm_tagged_pingpong", "short",
                               completion_semantic, memory_type_bi_dir, message_size,
                               fabric="efa")

@pytest.mark.parametrize("iteration_type",
                         [pytest.param("short", marks=pytest.mark.short),
                          pytest.param("standard", marks=pytest.mark.standard)])
def test_rdm_tagged_bw(cmdline_args, iteration_type, completion_semantic, memory_type, completion_type):
    command = "fi_rdm_tagged_bw"  + " " + perf_progress_model_cli
    efa_run_client_server_test(cmdline_args, command, iteration_type,
                               completion_semantic, memory_type, "all", completion_type=completion_type,
                               fabric="efa")

@pytest.mark.functional
def test_rdm_tagged_bw_range(cmdline_args, completion_semantic, memory_type, message_size):
    efa_run_client_server_test(cmdline_args, "fi_rdm_tagged_bw", "short",
                               completion_semantic, memory_type, message_size, fabric="efa")

@pytest.mark.functional
def test_rdm_tagged_bw_no_inject_range(cmdline_args, completion_semantic, inject_message_size):
    efa_run_client_server_test(cmdline_args, "fi_rdm_tagged_bw -j 0", "short",
                               completion_semantic, "host_to_host", inject_message_size, fabric="efa")

@pytest.mark.functional
@pytest.mark.parametrize("env_vars", [["FI_EFA_TX_SIZE=64"], ["FI_EFA_RX_SIZE=64"], ["FI_EFA_TX_SIZE=64", "FI_EFA_RX_SIZE=64"]])
def test_rdm_tagged_bw_small_tx_rx(cmdline_args, completion_semantic, memory_type, completion_type, env_vars):
    cmdline_args_copy = copy.copy(cmdline_args)
    for env_var in env_vars:
        cmdline_args_copy.append_environ(env_var)
    # Use a window size larger than tx/rx size
    efa_run_client_server_test(cmdline_args_copy, "fi_rdm_tagged_bw -W 128", "short",
                               completion_semantic, memory_type, "all", completion_type=completion_type,
                               fabric="efa")

@pytest.mark.functional
def test_rdm_tagged_bw_small_tx_rx(cmdline_args, completion_semantic, memory_type, completion_type):
    cmdline_args_copy = copy.copy(cmdline_args)
    cmdline_args_copy.append_environ("FI_EFA_RECVWIN_SIZE=1")
    # Use a window size larger than tx/rx size
    efa_run_client_server_test(cmdline_args_copy, "fi_rdm_tagged_bw -W 128", "short",
                               completion_semantic, memory_type, "all", completion_type=completion_type,
                               fabric="efa")

@pytest.mark.functional
def test_rdm_tagged_bw_use_fi_more(cmdline_args, completion_semantic, memory_type, message_size):
    efa_run_client_server_test(cmdline_args, "fi_rdm_tagged_bw --use-fi-more",
                               "short", completion_semantic, memory_type, message_size, fabric="efa")

# efa-direct does not support atomic
@pytest.mark.parametrize("iteration_type",
                         [pytest.param("short", marks=pytest.mark.short),
                          pytest.param("standard", marks=pytest.mark.standard)])
def test_rdm_atomic(cmdline_args, iteration_type, completion_semantic, memory_type):
    from copy import copy

    from common import ClientServerTest

    if "neuron" in memory_type:
        pytest.skip("Neuron does not fully support atomics")

    # the rdm_atomic test's run time has a high variance when running single c6gn instance.
    # the issue is tracked in:  https://github.com/ofiwg/libfabric/issues/7002
    # to mitigate the issue, set the maximum timeout of fi_rdm_atomic to 1800 seconds.
    cmdline_args_copy = copy(cmdline_args)
    command = "fi_rdm_atomic"  + " " + perf_progress_model_cli
    test = ClientServerTest(cmdline_args_copy, "fi_rdm_atomic", iteration_type, completion_semantic,
                            memory_type=memory_type, timeout=1800, fabric="efa")
    test.run()

@pytest.mark.functional
def test_rdm_tagged_peek(cmdline_args):
    from copy import copy

    from common import ClientServerTest

    test = ClientServerTest(cmdline_args, "fi_rdm_tagged_peek", timeout=1800)
    test.run()

# This test is run in serial mode because it takes a lot of memory
# It is skipped for efa-direct because 1GB exceeds device max msg size
@pytest.mark.serial
@pytest.mark.functional
def test_rdm_pingpong_1G(cmdline_args, completion_semantic):
    # Default window size is 64 resulting in 128GB being registered, which
    # exceeds max number of registered host pages
    efa_run_client_server_test(cmdline_args, "fi_rdm_pingpong -W 1", 2,
                               completion_semantic=completion_semantic, message_size=1073741824,
                               memory_type="host_to_host", warmup_iteration_type=0, fabric="efa")

@pytest.mark.functional
def test_rdm_pingpong_zcpy_recv(cmdline_args, memory_type_bi_dir, zcpy_recv_max_msg_size, zcpy_recv_message_size):
    if cmdline_args.server_id == cmdline_args.client_id:
        pytest.skip("no zero copy recv for intra-node communication")
    cmdline_args_copy = copy.copy(cmdline_args)
    cmdline_args_copy.append_environ("FI_EFA_ENABLE_SHM_TRANSFER=0")
    efa_run_client_server_test(cmdline_args_copy, f"fi_rdm_pingpong --max-msg-size {zcpy_recv_max_msg_size}",
                               "short", "transmit_complete", memory_type_bi_dir, zcpy_recv_message_size, fabric="efa")

@pytest.mark.functional
def test_rdm_bw_zcpy_recv(cmdline_args, memory_type, zcpy_recv_max_msg_size, zcpy_recv_message_size):
    if cmdline_args.server_id == cmdline_args.client_id:
        pytest.skip("no zero copy recv for intra-node communication")
    cmdline_args_copy = copy.copy(cmdline_args)
    cmdline_args_copy.append_environ("FI_EFA_ENABLE_SHM_TRANSFER=0")
    efa_run_client_server_test(cmdline_args_copy, f"fi_rdm_bw --max-msg-size {zcpy_recv_max_msg_size}",
                               "short", "transmit_complete", memory_type, zcpy_recv_message_size, fabric="efa")

@pytest.mark.functional
def test_rdm_bw_zcpy_recv_use_fi_more(cmdline_args, memory_type, zcpy_recv_max_msg_size, zcpy_recv_message_size):
    if cmdline_args.server_id == cmdline_args.client_id:
        pytest.skip("no zero copy recv for intra-node communication")
    cmdline_args_copy = copy.copy(cmdline_args)
    cmdline_args_copy.append_environ("FI_EFA_ENABLE_SHM_TRANSFER=0")
    efa_run_client_server_test(cmdline_args_copy, f"fi_rdm_bw --use-fi-more --max-msg-size {zcpy_recv_max_msg_size}",
                               "short", "transmit_complete", memory_type, zcpy_recv_message_size, fabric="efa")

@pytest.mark.functional
@pytest.mark.parametrize("comp_method", ["sread", "fd"])
def test_rdm_pingpong_sread(cmdline_args, completion_semantic, memory_type_bi_dir,
                            direct_message_size, support_sread, comp_method, fabric):
    if not support_sread:
        pytest.skip("sread not supported by efa device.")
    additional_env = ''
    if fabric == "efa" and comp_method == "fd":
        if cmdline_args.server_id == cmdline_args.client_id:
            pytest.skip("FI_WAIT_FD not supported for EFA protocol with SHM enabled")
        additional_env = "FI_EFA_ENABLE_SHM_TRANSFER=0"
    efa_run_client_server_test(cmdline_args, f"fi_rdm_pingpong -c {comp_method}", "short",
                               completion_semantic, memory_type_bi_dir,
                               direct_message_size if fabric == "efa-direct" else "all",
                               fabric=fabric, additional_env=additional_env)


# These tests skip efa-direct because efa-direct does not
# do memory registrations on behalf of the application
@pytest.mark.functional
@pytest.mark.serial
def test_mr_exhaustion_rdm_pingpong(cmdline_args, completion_semantic):
    import os
    binpath = cmdline_args.binpath or ""
    if not os.path.exists(os.path.join(binpath, "fi_efa_exhaust_mr_reg_rdm_pingpong")):
        pytest.skip("fi_efa_exhaust_mr_reg_rdm_pingpong requires efadv")
    efa_run_client_server_test(cmdline_args, "fi_efa_exhaust_mr_reg_rdm_pingpong", "short",
                               completion_semantic, "host_to_host", "all", timeout=1000,
                               fabric="efa")

@pytest.mark.parametrize("mr_cache", [True, False])
@pytest.mark.parametrize("iteration_type",
                         [pytest.param("short", marks=pytest.mark.short),
                          pytest.param("standard", marks=pytest.mark.standard)])
def test_rdm_pingpong_no_mr_local(cmdline_args, iteration_type, completion_semantic,
                      memory_type_bi_dir, completion_type, mr_cache):
    command = "fi_rdm_pingpong -M mr_local"  + " " + perf_progress_model_cli

    additional_env = ''
    if not mr_cache:
        additional_env = 'FI_EFA_MR_CACHE_ENABLE=0'

    efa_run_client_server_test(cmdline_args, command, iteration_type,
                               completion_semantic, memory_type_bi_dir, "all",
                               completion_type=completion_type, fabric="efa",
                               additional_env=additional_env)

@pytest.mark.parametrize("mr_cache", [True, False])
@pytest.mark.parametrize("iteration_type",
                         [pytest.param("short", marks=pytest.mark.short),
                          pytest.param("standard", marks=pytest.mark.standard)])
def test_rma_pingpong_no_mr_local(cmdline_args, iteration_type, completion_semantic,
                      memory_type_bi_dir, mr_cache):
    command = "fi_rma_pingpong -o writedata -M mr_local"  + " " + perf_progress_model_cli

    additional_env = ''
    if not mr_cache:
        additional_env = 'FI_EFA_MR_CACHE_ENABLE=0'

    efa_run_client_server_test(cmdline_args, command, iteration_type,
                               completion_semantic, memory_type_bi_dir, "all",
                               completion_type="queue", fabric="efa",
                               additional_env=additional_env)


@pytest.mark.functional
@pytest.mark.serial
@pytest.mark.parametrize("unexpected_path", [True, False])
# TODO: Add test with (1) larger message size that uses the long read protocol
# and (2) SHM provider with CMA protocol after fixing the test to poll CQ while
# waiting for the server to get send completions
@pytest.mark.parametrize("msg_size", [1, 1024, 8192, 24756])
# @pytest.mark.parametrize("msg_count", [1, 1024, 2048]) # below and above shm's default rx size
def test_implicit_av(cmdline_args, unexpected_path, msg_size):
    import os
    binpath = cmdline_args.binpath or ""
    if not os.path.exists(os.path.join(binpath, "fi_efa_implicit_av_test")):
        pytest.skip("implicit AV test not found")

    if (cmdline_args.server_id == cmdline_args.client_id) and unexpected_path:
        pytest.skip("SHM provider is not guaranteed to generate TX completions without posted receives")
    test_cmd = f"fi_efa_implicit_av_test -L -c 5 -S {msg_size}"
    if unexpected_path:
        test_cmd += " -X"

    efa_run_client_server_test(cmdline_args, test_cmd, "short",
                               "transmit_complete", "host_to_host", "all", fabric="efa")

@pytest.mark.functional
@pytest.mark.serial
# TODO: Add test with larger message size that uses the long read protocol after
# fixing the test to poll CQ while waiting for the server to get send completions
@pytest.mark.parametrize("msg_size", [1, 1024, 8192, 24756])
def test_implicit_av_limited_av_size_expected_path(cmdline_args, msg_size):
    import os
    binpath = cmdline_args.binpath or ""
    if not os.path.exists(os.path.join(binpath, "fi_efa_implicit_av_test")):
        pytest.skip("implicit AV test not found")

    additional_env = 'FI_EFA_IMPLICIT_AV_SIZE=3'

    test_cmd = f"fi_efa_implicit_av_test -L -c 5 -S {msg_size}"

    efa_run_client_server_test(cmdline_args, test_cmd, "short",
                               "transmit_complete", "host_to_host", "all", fabric="efa",
                               additional_env=additional_env, timeout=30)

@pytest.mark.functional
@pytest.mark.serial
# This test is expected to fail because the implicit AV size of 3 is smaller than
# the number of server endpoints of 5. So implicit AV entries will get evicted
# on the client and cause the test to hang.
@pytest.mark.xfail(strict=True)
# TODO: Add test with larger message size that uses the long read protocol after
# fixing the test to poll CQ while waiting for the server to get send completions
@pytest.mark.parametrize("msg_size", [1, 1024, 8192, 24756])
def test_implicit_av_limited_av_size_unexpected_path(cmdline_args, msg_size):
    import os
    binpath = cmdline_args.binpath or ""
    if not os.path.exists(os.path.join(binpath, "fi_efa_implicit_av_test")):
        pytest.skip("implicit AV test not found")

    if cmdline_args.server_id == cmdline_args.client_id:
        pytest.skip("This test will only fail for the EFA protocol path - which requires 2 nodes.")

    additional_env = 'FI_EFA_IMPLICIT_AV_SIZE=3'

    test_cmd = f"fi_efa_implicit_av_test -L -X -c 5 -S {msg_size}"

    efa_run_client_server_test(cmdline_args, test_cmd, "short",
                               "transmit_complete", "host_to_host", "all", fabric="efa",
                               additional_env=additional_env, timeout=15)

@pytest.mark.functional
def test_rmd_tagged_pingpong_truncate_error(cmdline_args, completion_semantic):
    if cmdline_args.server_id == cmdline_args.client_id:
        pytest.skip("client hits an assertion in debug mode in case of intra-node")
    executable = {"server": "fi_rdm_tagged_pingpong --timeout 1",
                  "client": "fi_rdm_tagged_pingpong --expect-error 265"}
    message_size = {"server": 2000001,
                    "client": 2000000}
    might_fail = {"server": True,
                  "client": False}

    efa_run_client_server_test(cmdline_args, executable, "short",
                               completion_semantic, "host_to_host",
                               message_size=message_size, fabric="efa",
                               timeout=15, might_fail=might_fail)



import pytest
from shm.shm_common import shm_run_client_server_test

SHM_DEFAULT_MAX_INJECT_SIZE = 4096
SHM_DEFAULT_RX_SIZE = 1024


@pytest.mark.functional
@pytest.mark.parametrize("msg_size", [1, 512, 9000]) # cover various switch points of shm protocols
@pytest.mark.parametrize("msg_count", [1, 1024, 2048]) # below and above shm's default rx size
def test_unexpected_msg(cmdline_args, msg_size, msg_count, memory_type, completion_semantic):
    from common import ClientServerTest
    if (msg_size > SHM_DEFAULT_MAX_INJECT_SIZE or memory_type != "host_to_host" or completion_semantic == "delivery_complete") and msg_count > SHM_DEFAULT_RX_SIZE:
        pytest.skip("SHM's CMA/IPC protocol currently cannot handle > rx size number of unexpected messages")
    shm_run_client_server_test(cmdline_args, f"fi_unexpected_msg -e rdm -M {msg_count}", iteration_type="short",
                               completion_semantic=completion_semantic, memory_type=memory_type,
                               message_size=msg_size, completion_type="queue", timeout=1800)


#!/usr/bin/env python3
# SPDX-License-Identifier: BSD-2-Clause OR GPL-2.0-only
# Copyright (c) 2021 Hewlett Packard Enterprise Development LP
help = f'''
Standalone REST server for local testing

TARGET /test
    Provides basic targets for GET, PUT, POST, PATCH, and DELETE.
    "Content-Type: application/json" header should be specified.
    Result is JSON data identifying the operation, and the supplied data.
    If the supplied data contains a JSON tag named 'return_code',
        the corresponding value will be used as the return code of the
        response.
    Exercise using ./curltest --auto

If --host is omitted, host is http://127.0.0.1 local address (if available)
If --host is 0.0.0.0, host is the current IP address of the node
'''

import argparse
import textwrap
import sys
import json

from argparse import ArgumentParser, HelpFormatter
from flask import Flask, request
from flask_restful import Api, Resource

class RawFormatter(HelpFormatter):
    def _fill_text(self, text, width, indent):
        return "\n".join([textwrap.fill(line, width) for line in textwrap.indent(textwrap.dedent(text), indent).splitlines()])

# Test code for CURL regression test
class selftestResource(Resource):
    def return_code(self, json):
        if json is not None and "return_code" in json:
            return json["return_code"]
        return 200

    def get(self):
        info = {
            'operation': 'GET',
            'data': ''
        }
        return info, self.return_code(None)

    def put(self):
        info = {
            'operation': 'PUT',
            'data': request.json
        }
        return info, self.return_code(request.json)

    def post(self):
        info = {
            'operation': 'POST',
            'data': request.json
        }
        return info, self.return_code(request.json)

    def patch(self):
        info = {
            'operation': 'PATCH',
            'data': request.json
        }
        return info, self.return_code(request.json)

    def delete(self):
        info = {
            'operation': 'DELETE',
            'data': request.json
        }
        return info, self.return_code(request.json)

def main(argv):
    parser = argparse.ArgumentParser(
        description=help, formatter_class=RawFormatter)
    parser.add_argument('--host', default=None)
    parser.add_argument('--port', default=None)
    args = parser.parse_args()

    app = Flask(__name__)
    api = Api(app);
    api.add_resource(selftestResource, '/test')
    app.run(debug=True, host=args.host, port=args.port)

if __name__ == "__main__":
    main(sys.argv[1:])



import os
import subprocess
import functools
import re
from enum import IntEnum
from collections import deque
from common import SshConnectionError, is_ssh_connection_error, has_ssh_connection_err_msg, ClientServerTest
from retrying import retry


@functools.lru_cache(2)
@retry(retry_on_exception=is_ssh_connection_error, stop_max_attempt_number=3, wait_fixed=5000)
def parse_lspci_tree(server_id):
    """
    Function parse the lspci output and construct a tree of the PCIe hierarchy

    Snippet of lspci -tv output on p5en instance type

    $ lspci -tv
    +-[0000:44]---00.0-[45-5c]----00.0-[46-5c]--+-00.0-[47]----00.0  Amazon.com, Inc. Elastic Network Adapter (ENA)
    |                                           +-00.1-[48]----00.0  Amazon.com, Inc. Elastic Network Adapter (ENA)
    |                                           +-00.2-[49]----00.0  Amazon.com, Inc. Elastic Network Adapter (ENA)
    |                                           +-00.3-[4a]----00.0  Amazon.com, Inc. Elastic Network Adapter (ENA)
    |                                           +-00.4-[4b]--
    |                                           +-00.5-[4c]--
    |                                           +-00.6-[4d]--
    |                                           +-00.7-[4e]--
    |                                           +-01.0-[4f]--
    |                                           +-01.1-[50]--
    |                                           +-01.2-[51]--
    |                                           +-01.3-[52]--
    |                                           +-01.4-[53]--
    |                                           +-01.5-[54]--
    |                                           +-01.6-[55]----00.0  Amazon.com, Inc. Elastic Fabric Adapter (EFA)
    |                                           +-01.7-[56]----00.0  Amazon.com, Inc. Elastic Fabric Adapter (EFA)
    |                                           +-02.0-[57]----00.0  Amazon.com, Inc. Elastic Fabric Adapter (EFA)
    |                                           +-02.1-[58]----00.0  Amazon.com, Inc. Elastic Fabric Adapter (EFA)
    |                                           +-02.2-[59]----00.0  NVIDIA Corporation GH100 [H200 SXM 141GB]
    |                                           +-02.3-[5a]----00.0  NVIDIA Corporation GH100 [H200 SXM 141GB]
    |                                           +-02.4-[5b]----00.0  Amazon.com, Inc. NVMe SSD Controller
    |                                           \-02.5-[5c]----00.0  Amazon.com, Inc. NVMe SSD Controller

    In the output above, "[0000:44]" is a bus downstream of the root complex
    "00.0-[45-5c]" is a bridge and "00.0-[46-5c]" is a switch upstream of buses 0x45 to 0x5c and 0x46 to 0x5c respectively

    Looking at the line with the EFA NIC,
    "01.6-[55]" is the bridge on bus 1 device 6 on bus 46 and is upstream of bus 55
    "00.0  Amazon.com, Inc. Elastic Fabric Adapter (EFA)" is the EFA NIC on bus 55, device function 00.0
    """
    timeout = 60
    result = subprocess.run([f'ssh {server_id}', 'lspci', '-tv'],
            shell=True,
            stdout=subprocess.PIPE, stderr=subprocess.PIPE,
            encoding="utf-8", timeout=timeout)

    if has_ssh_connection_err_msg(result.stderr):
        raise SshConnectionError()

    lines = result.stdout.strip().split('\n')

    tree = {}
    indent_stack = []

    for line in lines:
        indent = len(line) - len(line.lstrip(' '))

        bridge_matches = list(re.finditer(r'([0-9a-f]{2}\.[0-9a-f])-\[([0-9a-f]{2,4})-[0-9a-f]+\]', line))

        if bridge_matches:
            while indent_stack and indent_stack[-1][0] >= indent:
                indent_stack.pop()

            for bridge_match in bridge_matches:
                dev_func = bridge_match.group(1)
                bus_start = bridge_match.group(2).zfill(2)
                bridge_id = f'bridge_{bus_start}_{dev_func}'

                parent = indent_stack[-1][1] if indent_stack else None
                tree[bridge_id] = {'parent': parent, 'children': [], 'is_efa': False, 'slot': None}

                if parent:
                    if parent not in tree:
                        tree[parent] = {'parent': None, 'children': [], 'is_efa': False, 'slot': None}
                    tree[parent]['children'].append(bridge_id)

                indent_stack.append((indent, bridge_id))

        leaf_match = re.search(r'[+\\]-([0-9a-f]{2}\.[0-9a-f])-\[([0-9a-f]{2,4})\]----([0-9a-f]{2}\.[0-9a-f])', line)
        if leaf_match:
            slot = leaf_match.group(1)
            bus = leaf_match.group(2).zfill(2)
            dev_func = leaf_match.group(3)
            bdf = f'0000:{bus}:{dev_func}'
            is_efa = 'Amazon' in line and ('EFA' in line or 'efa' in line)

            parent = indent_stack[-1][1] if indent_stack else None
            tree[bdf] = {'parent': parent, 'children': [], 'is_efa': is_efa, 'slot': slot}

            if parent:
                if parent not in tree:
                    tree[parent] = {'parent': None, 'children': [], 'is_efa': False, 'slot': None}
                tree[parent]['children'].append(bdf)

    return tree

class CudaMemorySupport(IntEnum):
    NOT_INITIALIZED = -1
    NOT_SUPPORTED = 0
    DMA_BUF_ONLY = 1
    GDR_ONLY = 2
    DMABUF_GDR_BOTH = 3

    def __str__(self):
        return self.name

@retry(retry_on_exception=is_ssh_connection_error, stop_max_attempt_number=3, wait_fixed=5000)
def get_cuda_memory_support(cmdline_args, ip):
    """
    Execute check_dmabuf binary to determine CUDA memory support capabilities.
    
    Args:
        cmdline_args: Command line arguments containing binpath, timeout, provider, and environments.
        ip: IP address or hostname of the target machine.
        
    Returns:
        CudaMemorySupport: Enum value indicating hardware CUDA memory support type.
        
    Notes:
        - Executes check_dmabuf binary remotely via SSH with timeout
        - Maps return code directly to CudaMemorySupport enum values
        - Returns UNKNOWN for negative return codes indicating errors
        - Retries on SSH connection errors up to 3 times
    """
    binpath = cmdline_args.binpath or ""
    cmd = "timeout " + str(cmdline_args.timeout) \
          + " " + os.path.join(binpath, "cuda_check_dmabuf") \
          + " -p " + cmdline_args.provider
    if cmdline_args.environments:
        cmd = cmdline_args.environments + " " + cmd

    proc = subprocess.run("ssh {} {}".format(ip, cmd),
               stdout=subprocess.PIPE,
               stderr=subprocess.STDOUT,
               shell=True,
               universal_newlines=True)

    if has_ssh_connection_err_msg(proc.stdout):
        raise SshConnectionError()

    if proc.returncode < 0:
        return CudaMemorySupport.NOT_SUPPORTED

    print(f"The ssh return is {proc}")
    rc = proc.returncode
    if rc not in (CudaMemorySupport.NOT_SUPPORTED,
                  CudaMemorySupport.DMA_BUF_ONLY,
                  CudaMemorySupport.GDR_ONLY,
                  CudaMemorySupport.DMABUF_GDR_BOTH):
        print(f"[warn] check_dmabuf returned unexpected code {rc}, treating as NOT_INITIALIZED")
        return CudaMemorySupport.NOT_INITIALIZED

    return CudaMemorySupport(rc)

def efa_run_client_server_test(cmdline_args, executable, iteration_type,
                               completion_semantic, memory_type, message_size,
                               warmup_iteration_type=None, timeout=None,
                               completion_type="queue", fabric=None,
                               additional_env='',
                               might_fail=False):
    if timeout is None:
        timeout = cmdline_args.timeout

    # It is observed that cuda tests requires larger time-out limit to test all
    # message sizes (especailly when running with multiple workers).
    if "cuda" in memory_type:
        timeout = max(1000, timeout)

    # For host only test, we do not need to initialize non-system hmem ifaces
    if memory_type == "host_to_host":
        additional_env += ' FI_HMEM=system'

    test = ClientServerTest(cmdline_args, executable, iteration_type,
                            completion_semantic=completion_semantic,
                            datacheck_type="with_datacheck",
                            message_size=message_size,
                            memory_type=memory_type,
                            timeout=timeout,
                            warmup_iteration_type=warmup_iteration_type,
                            completion_type=completion_type, fabric=fabric,
                            additional_env=additional_env,
                            might_fail=might_fail)
    test.run()

@retry(retry_on_exception=is_ssh_connection_error, stop_max_attempt_number=3, wait_fixed=5000)
def efa_retrieve_hw_counter_value(hostname, hw_counter_name, efa_device_name=None):
    """
    retrieve the value of EFA's hardware counter
    hostname: a host that has efa
    hw_counter_name: EFA hardware counter name. Options are: lifespan, rdma_read_resp_bytes, rdma_read_wrs,recv_wrs,
                     rx_drops, send_bytes, tx_bytes, rdma_read_bytes,  rdma_read_wr_err, recv_bytes, rx_bytes, rx_pkts, send_wrs, tx_pkts
    efa_device_name: Name of the EFA device. Corresponds to the name of the EFA device's directory
    return: an integer that is sum of all EFA device's counter
    """

    if efa_device_name:
        efa_device_dir = efa_device_name
    else:
        efa_device_dir = '*'

    command = 'ssh {} cat "/sys/class/infiniband/{}/ports/*/hw_counters/{}"'.format(hostname, efa_device_dir, hw_counter_name)
    process = subprocess.run(command, shell=True, check=False, stdout=subprocess.PIPE, stderr=subprocess.PIPE, encoding="utf-8")
    if process.returncode != 0:
        if process.stderr and has_ssh_connection_err_msg(process.stderr):
            print("encountered ssh connection issue")
            raise SshConnectionError()
        # this can happen when OS is using older version of EFA kernel module
        return None

    linelist = process.stdout.split()
    sumvalue = 0
    for strvalue in linelist:
        sumvalue += int(strvalue)
    return sumvalue

def has_gdrcopy(hostname):
    """
    determine whether a host has gdrcopy installed
    hostname: a host
    return: a boolean
    """
    command = "ssh {} /bin/bash --login -c lsmod | grep gdrdrv".format(hostname)
    process = subprocess.run(command, shell=True, check=False, stdout=subprocess.PIPE)
    return process.returncode == 0

@retry(retry_on_exception=is_ssh_connection_error, stop_max_attempt_number=3, wait_fixed=5000)
def has_rdma(cmdline_args, operation):
    """
    determine whether both client/server sides have rdma <operation> enabled in efa device
    cmdline_args: command line argument object, including server/client id, and timeout
    operation: rdma operation name, allowed values are read and write
    return: a boolean
    """
    assert operation in ["read", "write", "writedata"]
    binpath = cmdline_args.binpath or ""
    cmd = "timeout " + str(cmdline_args.timeout) \
          + " " + os.path.join(binpath, f"fi_efa_rdma_checker -o {operation}")
    if cmdline_args.environments:
        cmd = cmdline_args.environments + " " + cmd
    server_proc = subprocess.run("ssh {} {}".format(cmdline_args.server_id, cmd),
               stdout=subprocess.PIPE,
               stderr=subprocess.STDOUT,
               shell=True,
               universal_newlines=True)
    if has_ssh_connection_err_msg(server_proc.stdout):
        raise SshConnectionError()

    client_proc = subprocess.run("ssh {} {}".format(cmdline_args.client_id, cmd),
               stdout=subprocess.PIPE,
               stderr=subprocess.STDOUT,
               shell=True,
               universal_newlines=True)
    if has_ssh_connection_err_msg(client_proc.stdout):
        raise SshConnectionError()

    return (server_proc.returncode == 0 and client_proc.returncode == 0)

def efa_retrieve_gid(hostname):
    """
    return the GID of efa device on a host
    hostname: a host
    return: a string if the host has efa device,
            None otherwise
    """
    command = "ssh {} ibv_devinfo  -v | grep GID | awk '{{print $NF}}' | head -n 1".format(hostname)
    try:
        process = subprocess.run(command, shell=True, check=True, stdout=subprocess.PIPE)
    except subprocess.CalledProcessError:
        # this can happen on instance without EFA device
        return None

    return process.stdout.decode("utf-8").strip()

@functools.lru_cache(10)
@retry(retry_on_exception=is_ssh_connection_error, stop_max_attempt_number=3, wait_fixed=5000)
def get_efa_device_names(server_id):
    timeout = 60

    # This command returns a list of EFA devices names
    command = "ssh {} 'fi_info -p efa -t FI_EP_RDM -f efa || /opt/amazon/efa/bin/fi_info -p efa -t FI_EP_RDM -f efa' | grep domain".format(server_id)
    proc = subprocess.run(command, shell=True,
                          stdout=subprocess.PIPE, stderr=subprocess.PIPE,
                          encoding="utf-8", timeout=timeout)

    if has_ssh_connection_err_msg(proc.stderr):
        raise SshConnectionError()

    devices = []
    stdouts = proc.stdout.strip().split("\n")
    #
    # Example out of fi_info -p efa -t FI_EP_RDM -f efa | grep domain are like the following:
    #     domain: rdmap0s31-rdm
    #     ...
    #
    # Extract the device name from the second column of the stdout
    # by removing the -rdm suffix
    for line in stdouts:
        parts = line.split()
        if len(parts) > 1 and "rdm" in parts[1]:
            devices.append(parts[1].replace("-rdm", ""))
    return devices

@functools.lru_cache(16)
@retry(retry_on_exception=is_ssh_connection_error, stop_max_attempt_number=3, wait_fixed=5000)
def get_gpu_bdf(server_id, gpu_index):
    timeout = 60

    result = subprocess.run(
        [f'ssh {server_id}', 'nvidia-smi', '--query-gpu=pci.bus_id', '--format=csv,noheader', '--id', str(gpu_index)],
            shell=True,
            stdout=subprocess.PIPE, stderr=subprocess.PIPE,
            encoding="utf-8", timeout=timeout)

    if has_ssh_connection_err_msg(result.stderr):
        raise SshConnectionError()

    bdf_raw = result.stdout.strip().lower()
    bdf_parts = bdf_raw.split(':')
    if len(bdf_parts) != 3:
        return None
    domain = bdf_parts[0]
    if len(domain) < 4:
        domain = domain.zfill(4)
    else:
        domain = domain[-4:].zfill(4)
    bus = bdf_parts[1].zfill(2)
    dev_func = bdf_parts[2]
    return f'{domain}:{bus}:{dev_func}'

@functools.lru_cache(32)
@retry(retry_on_exception=is_ssh_connection_error, stop_max_attempt_number=3, wait_fixed=5000)
def get_rdma_core_name_for_efa_nic(server_id, bdf):
    timeout = 60
    result = subprocess.run(
        [f'ssh {server_id}', 'ls', f'/sys/bus/pci/devices/{bdf}/infiniband'],
        shell=True,
        stdout=subprocess.PIPE, stderr=subprocess.PIPE,
        encoding="utf-8", timeout=timeout)

    if has_ssh_connection_err_msg(result.stderr):
        raise SshConnectionError()

    if result.returncode == 0 and result.stdout.strip():
        return result.stdout.strip().split()[0]
    return None

def get_closest_efa_nics_for_gpu(server_id, gpu_index, pcie_tree):
    """
    BFS traversal to find the closest EFA NICs for a given GPU
    """
    gpu_bdf = get_gpu_bdf(server_id, gpu_index)

    if gpu_bdf not in pcie_tree:
        return []

    visited = set()
    queue = deque([(gpu_bdf, 0)])
    closest_distance = None
    closest_efas = []

    while queue:
        node, distance = queue.popleft()

        if node in visited:
            continue
        visited.add(node)

        if closest_distance is not None and distance > closest_distance:
            break

        if node in pcie_tree:
            if pcie_tree[node]['is_efa'] and node != gpu_bdf:
                rdma_name = get_rdma_core_name_for_efa_nic(server_id, node)
                if rdma_name:
                    if closest_distance is None:
                        closest_distance = distance
                    if distance == closest_distance:
                        closest_efas.append(rdma_name)

            # Add parent
            if pcie_tree[node]['parent'] and pcie_tree[node]['parent'] not in visited:
                queue.append((pcie_tree[node]['parent'], distance + 1))

            # Add children
            for child in pcie_tree[node]['children']:
                if child not in visited:
                    queue.append((child, distance + 1))

    return closest_efas

def get_efa_device_name_for_hmem_device(ip, hmem_device_id, num_hmem_devices):
    # this function implemented a simple way to find the closest EFA device for a given
    # hmem device. It assumes EFA devices names are in order (which is usually true but not always)
    #
    # For example, one a system with 8 accelerators and 4 EFA devices, this function would
    # for accelerator 0 and 1, return EFA device 0
    # for accelerator 2 and 3, return EFA device 1
    # for accelerator 4 and 5, return EFA device 2
    # for accelerator 6 and 7, return EFA device 3
    efa_devices = get_efa_device_names(ip)
    num_efa = len(efa_devices)
    return efa_devices[(hmem_device_id * num_efa) // num_hmem_devices]

@functools.lru_cache(10)
def get_efa_device_name_for_cuda_device(ip, cuda_device_id, num_cuda_devices):
    """
    Traverse the PCIe hierarchy and find the closest EFA NIC for a given GPU
    If the PCIe hierarchy traversal fails, fallback to a simple round robin
    """
    efa_devices = get_efa_device_names(ip)
    num_efa = len(efa_devices)

    pcie_tree = parse_lspci_tree(ip)
    closest_nics = get_closest_efa_nics_for_gpu(ip, cuda_device_id, pcie_tree)

    if not closest_nics:
        return get_efa_device_name_for_hmem_device(ip, cuda_device_id, num_cuda_devices)

    # Each GPU can use a different EFA NIC
    if num_efa >= num_cuda_devices:
        return closest_nics[0]

    # Multiple GPUs share the same EFA NIC
    return closest_nics[(cuda_device_id * num_efa) // num_cuda_devices]

@retry(retry_on_exception=is_ssh_connection_error, stop_max_attempt_number=3, wait_fixed=5000)
def support_cq_interrupts(hostname):
    """
    determine whether an EFA device supports fi_cq_sread
    This is based on checking if the device has more than 1 completion vector,
    which indicates support for CQ interrupts
    hostname: a host that has efa
    return: True if sread is supported, False otherwise, None if unable to determine
    """
    try:
        efa_devices = get_efa_device_names(hostname)
        if not efa_devices:
            return None
        efa_device_name = efa_devices[0]
    except Exception:
        return None

    command = 'ssh {} ibv_devinfo -d {} -v | grep "num_comp_vectors"'.format(hostname, efa_device_name)
    process = subprocess.run(command, shell=True, check=False, stdout=subprocess.PIPE, stderr=subprocess.PIPE, encoding="utf-8")

    if process.returncode != 0:
        if process.stderr and has_ssh_connection_err_msg(process.stderr):
            print("encountered ssh connection issue")
            raise SshConnectionError()
        return None

    # Example output of ibv_devinfo -d rdmap85s0 -v | grep "num_comp_vectors"
    # num_comp_vectors:		32
    lines = process.stdout.split('\n')
    for line in lines:
        line = line.strip()
        if 'num_comp_vectors' in line:
            parts = line.split(':')
            if len(parts) >= 2:
                try:
                    comp_vectors = int(parts[1].strip())
                    # Support sread if num_comp_vectors > 1
                    return comp_vectors > 1
                except ValueError:
                    continue

    return None



import pytest

@pytest.mark.functional
def test_flood_peer(cmdline_args, fabric):
    from common import ClientServerTest
    test = ClientServerTest(cmdline_args, f"fi_flood -e rdm -W 6400 -S 512 -T 5",
                            timeout=300, fabric=fabric)
    test.run()



import pytest
from shm.shm_common import shm_run_client_server_test
from common import perf_progress_model_cli


@pytest.fixture(params=["r:4048,4,4148",
                        "r:8000,4,9000",
                        "r:17000,4,18000"])
def rma_pingpong_message_size(request):
    return request.param


@pytest.mark.parametrize("operation_type", ["writedata"])
@pytest.mark.parametrize("iteration_type",
                         [pytest.param("short", marks=pytest.mark.short),
                          pytest.param("standard", marks=pytest.mark.standard)])
def test_rma_pingpong(cmdline_args, iteration_type, operation_type, completion_semantic, memory_type):
    command = "fi_rma_pingpong -e rdm"
    command = command + " -o " + operation_type + " " + perf_progress_model_cli
    shm_run_client_server_test(cmdline_args, command, iteration_type, completion_semantic, memory_type, "all")

@pytest.mark.functional
@pytest.mark.parametrize("operation_type", ["writedata"])
def test_rma_pingpong_range(cmdline_args, operation_type, completion_semantic, rma_pingpong_message_size, memory_type):
    command = "fi_rma_pingpong -e rdm"
    command = command + " -o " + operation_type
    shm_run_client_server_test(cmdline_args, command, "short", completion_semantic, memory_type, rma_pingpong_message_size)



import os
import pytest
from common import ClientServerTest, has_cuda


@pytest.mark.short
@pytest.mark.functional
@pytest.mark.cuda_memory
def test_gda_send_recv(cmdline_args, direct_message_size, fabric):
    binpath = cmdline_args.binpath or ""
    if not os.path.exists(os.path.join(binpath, "fi_efa_gda")):
        pytest.skip("fi_efa_gda is not built")

    if not cmdline_args.do_dmabuf_reg_for_hmem:
        pytest.skip("DMABUF is required for GDA tests")

    if not has_cuda(cmdline_args.client_id) or not has_cuda(cmdline_args.server_id):
        pytest.skip("Client and server both need a cuda device")

    if fabric != "efa-direct":
        pytest.skip("GDA only works for efa-direct fabric")

    test = ClientServerTest(cmdline_args, "fi_efa_gda",
                            message_size=direct_message_size,
                            iteration_type="short",
                            memory_type="cuda_to_cuda",
                            fabric=fabric)
    test.run()



import pytest
import time
from efa_common import (
    has_rdma, 
    support_cq_interrupts,
    CudaMemorySupport,
    get_cuda_memory_support,
)

# The memory types for bi-directional tests.
memory_type_list_bi_dir = [
    pytest.param("host_to_host"),
    pytest.param("host_to_cuda", marks=pytest.mark.cuda_memory),
    pytest.param("cuda_to_cuda", marks=pytest.mark.cuda_memory),
    pytest.param("host_to_neuron", marks=pytest.mark.neuron_memory),
    pytest.param("neuron_to_neuron", marks=pytest.mark.neuron_memory),
    pytest.param("host_to_rocr", marks=pytest.mark.rocr_memory),
    pytest.param("rocr_to_rocr", marks=pytest.mark.rocr_memory),
]

# Add more memory types that are useful for uni-directional tests.
memory_type_list_all = memory_type_list_bi_dir + [
    pytest.param("cuda_to_host", marks=pytest.mark.cuda_memory),
    pytest.param("neuron_to_host", marks=pytest.mark.neuron_memory),
    pytest.param("rocr_to_host", marks=pytest.mark.rocr_memory),
]

hmem_type_list = [
    pytest.param("cuda", marks=pytest.mark.cuda_memory),
    pytest.param("neuron", marks=pytest.mark.neuron_memory),
]

@pytest.fixture(scope="module", params=hmem_type_list)
def hmem_type(request):
    return request.param

@pytest.fixture(scope="module", params=memory_type_list_all)
def memory_type(request):
    return request.param

@pytest.fixture(scope="module", params=memory_type_list_bi_dir)
def memory_type_bi_dir(request):
    return request.param

@pytest.fixture(scope="module", params=["read", "writedata", "write"])
def rma_operation_type(request):
    return request.param

@pytest.fixture(scope="module")
def rma_bw_memory_type(memory_type, rma_operation_type):
    is_test_bi_dir = False if rma_operation_type == "writedata" else True
    if is_test_bi_dir and (memory_type not in [_.values[0] for _ in memory_type_list_bi_dir]):
        pytest.skip("Duplicated memory type for bi-directional test")
    return memory_type

@pytest.fixture(scope="function")
def rma_bw_completion_semantic(cmdline_args, completion_semantic, rma_operation_type):
    if completion_semantic != 'delivery_complete':
        # There is no difference between DC and non-DC for read as it's
        # not a transmission
        if rma_operation_type == 'read':
            pytest.skip("Duplicate completion semantic for fi_read test")
        assert rma_operation_type in ['write', 'writedata']
        # If device support rdma write, all the transmissions are DC
        if has_rdma(cmdline_args, 'write'):
            pytest.skip("Duplicate completion semantic for fi_write* test")
    return completion_semantic


@pytest.fixture(scope="module", params=["r:0,4,64",
                                        "r:4048,4,4148",
                                        "r:8000,4,9000",
                                        "r:17000,4,18000",
                                        "r:0,4096,1048576"])
def message_size(request):
    return request.param


@pytest.fixture(scope="module", params=["r:0,4,64",
                                        "r:4048,4,4148",
                                        "r:8000,4,9000",])
def inject_message_size(request):
    return request.param


@pytest.fixture(scope="module", params=["r:0,4,32",
                                        "r:0,1024,8192",])
def zcpy_recv_message_size(request):
    return request.param

@pytest.fixture(scope="module")
def zcpy_recv_max_msg_size(request):
    return 8192

@pytest.fixture(scope="module", params=["r:0,4,32",
                                        "r:0,1024,8192",])
def direct_message_size(request):
    return request.param

# TODO: Include 0 byte test when we support 0 byte rma inject
@pytest.fixture(scope="module", params=["r:1,4,32",
                                        "r:1,1024,8192",])
def direct_rma_size(request):
    return request.param

@pytest.fixture(scope="module", params=["efa", "efa-direct"])
def fabric(request):
    return request.param

@pytest.fixture(scope="function")
def rma_fabric(cmdline_args, fabric):
    if fabric == 'efa-direct' and (
        not has_rdma(cmdline_args, 'read') or
        not has_rdma(cmdline_args, 'write') or
        not has_rdma(cmdline_args, 'writedata')
    ):
        pytest.skip("FI_RMA is not supported. Skip rma tests on efa-direct.")
    return fabric


@pytest.fixture(scope="function", params=["rx-cq-data", "no-rx-cq-data"])
def rx_cq_data_cli(request, fabric, rma_operation_type):
    if request.param == "no-rx-cq-data":
        if rma_operation_type != "writedata":
            pytest.skip("the rx cq data mode is only applied for writedata")
        if fabric == "efa-direct" :
            return " --no-rx-cq-data"
        else:
            pytest.skip("efa fabric ignores the rx cq data mode")
    return " "


def cuda_memory_type_validation(cmdline_args):
    """
    Validate CUDA memory type configuration against hardware capabilities at session startup.
    
    Args:
        cmdline_args: Command line arguments containing dmabuf configuration.
        
    Returns:
        None
        
    Notes:
        - Skips tests if user specified non-dmabuf but hardware only supports DMA_BUF_ONLY
        - Only validates if CUDA tests are being run
    """
    # Check if CUDA tests are being run via expression
    print("Running cuda_memory_type_validation() validation checks!")
    
    cuda_support: CudaMemorySupport = get_cuda_memory_support(
                                            cmdline_args=cmdline_args, 
                                            ip=cmdline_args.server_id
                                        )

    if cuda_support == CudaMemorySupport.NOT_INITIALIZED:
        pytest.fail("CUDA memory support never initialized")
    
    do_dmabuf = cmdline_args.do_dmabuf_reg_for_hmem
    
    print(f"Correctly defined dma buf mode {do_dmabuf} and return {cuda_support}!")
    
    return


@pytest.fixture(scope="function", autouse=True)
def cuda_validation_fixture(request, cmdline_args):
    """Auto-run CUDA validation if CUDA tests are present."""
    # Check if the current test has cuda_memory mark
    has_cuda_mark = any(mark.name == 'cuda_memory' for mark in request.node.iter_markers())
    
    if has_cuda_mark:
        cuda_memory_type_validation(cmdline_args)
    else:
        print("No CUDA memory mark, skipping validation")


@pytest.hookimpl(hookwrapper=True)
def pytest_collection_modifyitems(session, config, items):
    # Called after collection has been performed, may filter or re-order the items in-place
    # We use this hook to always run the MR exhaustion test at the end
    mr_exhaustion_tests, other_tests = [], []
    for item in items:
        if "mr_exhaustion" in item.name:
            mr_exhaustion_tests.append(item)
        else:
            other_tests.append(item)

    yield other_tests + mr_exhaustion_tests


@pytest.fixture(scope="function")
def support_sread(cmdline_args):
    """Check if both server and client support cq interrupts."""
    return (support_cq_interrupts(cmdline_args.server_id) and
            support_cq_interrupts(cmdline_args.client_id))





import pytest
from common import ClientServerTest

@pytest.mark.unstable
def test_multi_ep_stress_standard(cmdline_args):
    cmd = f"fi_efa_multi_ep_stress"
    test = ClientServerTest(cmdline_args, cmd, message_size=1024, fabric="efa", additional_env="FI_EFA_ENABLE_SHM_TRANSFER=0")
    test.run()

@pytest.mark.unstable
def test_multi_ep_stress_tagged(cmdline_args):
    cmd = f"fi_efa_multi_ep_stress --op-type tagged"
    test = ClientServerTest(cmdline_args, cmd, message_size=1024, fabric="efa", additional_env="FI_EFA_ENABLE_SHM_TRANSFER=0")
    test.run()

@pytest.mark.unstable
def test_multi_ep_stress_writedata(cmdline_args):
    cmd = f"fi_efa_multi_ep_stress --op-type writedata"
    test = ClientServerTest(cmdline_args, cmd, message_size=1024, fabric="efa", additional_env="FI_EFA_ENABLE_SHM_TRANSFER=0")
    test.run()

@pytest.mark.unstable
def test_multi_ep_stress_multi_receivers(cmdline_args):
    cmd = f"fi_efa_multi_ep_stress --receiver-workers 4"
    test = ClientServerTest(cmdline_args, cmd, message_size=1024, fabric="efa", additional_env="FI_EFA_ENABLE_SHM_TRANSFER=0")
    test.run()


@pytest.mark.unstable
def test_multi_ep_stress_multi_senders(cmdline_args):
    cmd = f"fi_efa_multi_ep_stress --sender-workers 10"
    test = ClientServerTest(cmdline_args, cmd, message_size=1024, fabric="efa", additional_env="FI_EFA_ENABLE_SHM_TRANSFER=0")
    test.run()

@pytest.mark.unstable
def test_multi_ep_stress_shared_cq(cmdline_args):
    cmd = f"fi_efa_multi_ep_stress --receiver-workers 4 --shared-cq"
    test = ClientServerTest(cmdline_args, cmd, message_size=1024, fabric="efa", additional_env="FI_EFA_ENABLE_SHM_TRANSFER=0")
    test.run()

@pytest.mark.unstable
def test_multi_ep_stress_shared_av(cmdline_args):
    cmd = f"fi_efa_multi_ep_stress --sender-workers 8 --receiver-workers 2 --shared-av"
    test = ClientServerTest(cmdline_args, cmd, message_size=1024, fabric="efa", additional_env="FI_EFA_ENABLE_SHM_TRANSFER=0")
    test.run()

@pytest.mark.unstable
def test_multi_ep_stress_shared_av_and_cq(cmdline_args):
    cmd = f"fi_efa_multi_ep_stress --sender-workers 8 --receiver-workers 2 --shared-av --shared-cq"
    test = ClientServerTest(cmdline_args, cmd, message_size=1024, fabric="efa", additional_env="FI_EFA_ENABLE_SHM_TRANSFER=0")
    test.run()

@pytest.mark.unstable
def test_multi_ep_stress_transient_client(cmdline_args):
    cmd = f"fi_efa_multi_ep_stress --sender-ep-cycles 5"
    test = ClientServerTest(cmdline_args, cmd, message_size=1024, fabric="efa", additional_env="FI_EFA_ENABLE_SHM_TRANSFER=0")
    test.run()

@pytest.mark.unstable
@pytest.mark.parametrize("configuration", [{"sender_workers": 2, "receiver_workers": 4, "sender_ep_cycles": 10, "receiver_ep_cycles": 20},
                                           {"sender_workers": 4, "receiver_workers": 2, "sender_ep_cycles": 20, "receiver_ep_cycles": 10},
                                           {"sender_workers": 2, "receiver_workers": 4, "sender_ep_cycles": 20, "receiver_ep_cycles": 10}
                                            ])
def test_multi_ep_stress_multi_and_transient_sender_receiver(cmdline_args, configuration):
    sender_workers = configuration["sender_workers"]
    receiver_workers = configuration["receiver_workers"]
    sender_ep_cycles = configuration["sender_ep_cycles"]
    receiver_ep_cycles = configuration["receiver_ep_cycles"]
    cmd = f"fi_efa_multi_ep_stress --sender-workers {sender_workers} --receiver-workers {receiver_workers} --sender-ep-cycles {sender_ep_cycles} --receiver-ep-cycles {receiver_ep_cycles}"
    test = ClientServerTest(cmdline_args, cmd, message_size=1024, fabric="efa", additional_env="FI_EFA_ENABLE_SHM_TRANSFER=0")
    test.run()



#!/usr/bin/env python3
# SPDX-License-Identifier: BSD-2-Clause OR GPL-2.0-only
# SPDX-FileCopyrightText: Copyright Amazon.com, Inc. or its affiliates. All rights reserved

import argparse
import builtins
import os
import sys

import yaml

import pytest
from junitparser import JUnitXml
from pytest import ExitCode


def get_option_longform(option_name, option_params):
    '''
        get the long form command line option name of an option
    '''
    return option_params.get("longform", "--" + option_name.replace("_", "-"))

def get_ubertest_test_type(fabtests_testsets):
    test_list = fabtests_testsets.split(",")

    for test in test_list:
        if test == "quick" or test == "ubertest_quick" or test == "ubertest":
            return "quick"

        if test == "all" or test == "ubertest_all":
            return "all"

        if test == "verify" or test == "ubertest_verify":
            return "verify"

    return None

def fabtests_testsets_to_pytest_markers(fabtests_testsets, run_mode=None):
    if run_mode:
        assert run_mode in ["serial", "parallel"]

    test_set = set()
    test_list = fabtests_testsets.split(",")

    # use set() to remove duplicate test set
    for test in test_list:
        if test == "quick":
            test_set.add("unit")
            test_set.add("functional")
            test_set.add("short")
            test_set.add("ubertest_quick")
        elif test =="ubertest":
            test_set.add("ubertest_quick")
        elif test == "all":
            test_set.add("unit")
            test_set.add("functional")
            test_set.add("standard")
            test_set.add("multinode")
            test_set.add("ubertest_all")
        elif test == "verify":
            test_set.add("ubertest_verify")
        else:
            test_set.add(test)

    markers = None
    for test in test_set:
        if markers is None:
            markers = test[:]
        else:
            markers += " or " + test

    if run_mode:
        if run_mode == "serial":
            markers = "(" + markers + ") and (serial)"
        else:
            assert run_mode == "parallel"
            markers = "(" + markers + ") and (not serial)"

    return markers

def get_default_exclusion_file(fabtests_args):
    test_configs_dir = os.path.abspath(os.path.join(get_pytest_root_dir(), "..", "test_configs"))
    exclusion_file = os.path.join(test_configs_dir, fabtests_args.provider,
                                  fabtests_args.provider + ".exclude")
    if not os.path.exists(exclusion_file):
        return None

    return exclusion_file

def get_default_ubertest_config_file(fabtests_args):
    test_configs_dir = os.path.abspath(os.path.join(get_pytest_root_dir(), "..", "test_configs"))
    provider = fabtests_args.provider
    if provider.find(";") != -1:
        core,util = fabtests_args.provider.split(";")
        cfg_file = os.path.join(test_configs_dir, util, core + ".test")
    else:
        core = fabtests_args.provider
        ubertest_test_type = get_ubertest_test_type(fabtests_args.testsets)
        if not ubertest_test_type:
            return None

        cfg_file = os.path.join(test_configs_dir, core, ubertest_test_type + ".test")

    if not os.path.exists(cfg_file):
        return None

    return cfg_file

def add_common_arguments(parser, shared_options):
    for option_name in shared_options.keys():
        option_params = shared_options[option_name]
        option_longform = get_option_longform(option_name, option_params)
        option_shortform = option_params.get("shortform")
        option_type = option_params["type"]
        option_helpmsg = option_params["help"]
        option_default = option_params.get("default")
        if option_type == "int" and not (option_default is None):
            option_default = int(option_default)

        if option_shortform:
            forms = [option_shortform, option_longform]
        else:
            forms = [option_longform]

        if option_type == "bool" or option_type == "boolean":
            parser.add_argument(*forms,
                                dest=option_name, action="store_true",
                                help=option_helpmsg, default=option_default)
        else:
            assert option_type == "str" or option_type == "int"
            parser.add_argument(*forms,
                                dest=option_name, type=getattr(builtins, option_type),
                                help=option_helpmsg, default=option_default)

def fabtests_args_to_pytest_args(fabtests_args, shared_options, run_mode):
    pytest_args = []

    if run_mode == "parallel":
        pytest_args.append("-n")
        pytest_args.append(str(fabtests_args.nworkers))

    pytest_args.append("--provider=" + fabtests_args.provider)
    pytest_args.append("--server-id=" + fabtests_args.server_id)
    pytest_args.append("--client-id=" + fabtests_args.client_id)

    # -v make pytest to print 1 line for each test
    pytest_args.append("-v")

    pytest_verbose_options = {
            0 : "-rN",      # print no extra information
            1 : "-rfE",     # print extra information for failed test(s)
            2 : "-rfEsx",   # print extra information for failed/skipped test(s)
            3 : "-rA",      # print extra information for all test(s) (failed/skipped/passed)
        }

    pytest_args.append(pytest_verbose_options[fabtests_args.verbose])

    verbose_fail = fabtests_args.verbose > 0
    if verbose_fail:
        # Use short python trace back because it show captured stdout of failed tests
        pytest_args.append("--tb=short")
    else:
        pytest_args.append("--tb=no")

    markers = fabtests_testsets_to_pytest_markers(fabtests_args.testsets, run_mode)
    pytest_args.append("-m")
    pytest_args.append(markers)

    if fabtests_args.expression:
        pytest_args.append("-k")
        pytest_args.append(fabtests_args.expression)

    if fabtests_args.html:
        pytest_args.append("--html")
        pytest_args.append(os.path.abspath(fabtests_args.html))
        pytest_args.append("--self-contained-html")

    if fabtests_args.junit_xml:
        pytest_args.append("--junit-xml")
        file_name = os.path.abspath(fabtests_args.junit_xml)
        if run_mode:
            file_name += "." + run_mode
        pytest_args.append(file_name)
        if fabtests_args.junit_logging:
            pytest_args.append("-o")
            pytest_args.append("junit_logging=" + fabtests_args.junit_logging)

    # add options shared between runfabtests.py and libfabric pytest
    for option_name in shared_options.keys():
        option_params = shared_options[option_name]
        option_longform = get_option_longform(option_name, option_params)
        option_type = option_params["type"]
 
        if not hasattr(fabtests_args, option_name):
            continue

        option_value = getattr(fabtests_args, option_name)
        if (option_value is None):
            continue

        if option_type == "bool" or option_type == "boolean":
            assert option_value
            pytest_args.append(option_longform)
        else:
            assert option_type == "str" or option_type == "int"
            pytest_args.append(option_longform + "=" + str(option_value))

    if not hasattr(fabtests_args, "exclusion_file") or not fabtests_args.exclusion_file:
        default_exclusion_file = get_default_exclusion_file(fabtests_args)
        if default_exclusion_file:
            pytest_args.append("--exclusion-file=" + default_exclusion_file)

    if not hasattr(fabtests_args, "ubertest_config_file") or not fabtests_args.ubertest_config_file:
        default_ubertest_config_file = get_default_ubertest_config_file(fabtests_args)
        if default_ubertest_config_file:
            pytest_args.append("--ubertest-config-file=" + default_ubertest_config_file)

    return pytest_args

def get_pytest_root_dir():
    '''
        find the pytest root directory according the location of runfabtests.py
    '''
    script_path = os.path.abspath(sys.argv[0])
    script_dir = os.path.dirname(script_path)
    if os.path.basename(script_dir) == "bin":
        # runfabtests.py is part of a fabtests installation
        pytest_root_dir = os.path.abspath(os.path.join(script_dir, "..", "share", "fabtests", "pytest"))
    elif os.path.basename(script_dir) == "scripts":
        # runfabtests.py is part of a fabtests source code
        pytest_root_dir = os.path.abspath(os.path.join(script_dir, "..", "pytest"))
    else:
        raise RuntimeError("Error: runfabtests.py is under directory {}, "
                "which is neither part of fabtests installation "
                "nor part of fabetsts source code".format(script_dir))

    if not os.path.exists(pytest_root_dir):
        raise RuntimeError("Deduced pytest root directory {} does not exist!".format(pytest_root_dir))

    return pytest_root_dir

def get_pytest_relative_case_dir(fabtests_args, pytest_root_dir):
    '''
        the directory that contains test cases, relative to pytest_root_dir
    '''
    # provider's own test directory (if exists) overrides default
    pytest_case_dir = os.path.join(pytest_root_dir, fabtests_args.provider)
    if os.path.exists(pytest_case_dir):
        return fabtests_args.provider

    assert os.path.exists(os.path.join(pytest_root_dir, "default"))
    return "default"


def run(fabtests_args, shared_options, run_mode):
    prev_cwd = os.getcwd()
    pytest_root_dir = get_pytest_root_dir()

    pytest_args = fabtests_args_to_pytest_args(fabtests_args, shared_options, run_mode)
    pytest_args.append(get_pytest_relative_case_dir(fabtests_args, pytest_root_dir))

    pytest_command = "cd " + pytest_root_dir + "; pytest"
    for arg in pytest_args:
        if arg.find(' ') != -1:
            arg = "'" + arg + "'"
        pytest_command += " " + arg
    print(pytest_command)

    # actually running tests

    os.chdir(pytest_root_dir)
    status = pytest.main(pytest_args)
    os.chdir(prev_cwd)
    return status


def main():
    pytest_root_dir = get_pytest_root_dir()

    # pytest/options.yaml contains the definition of a list of options that are
    # shared between runfabtests.py and pytest
    option_yaml = os.path.join(pytest_root_dir, "options.yaml")
    if not os.path.exists(option_yaml):
        print("Error: option definition yaml file {} not found!".format(option_yaml))
        exit(1)

    shared_options = yaml.safe_load(open(option_yaml))

    parser = argparse.ArgumentParser(description="libfabric integration test runner")

    parser.add_argument("provider", type=str, help="libfabric provider")
    parser.add_argument("server_id", type=str, help="server ip or hostname")
    parser.add_argument("client_id", type=str, help="client ip or hostname")
    parser.add_argument("-t", dest="testsets", type=str, default="quick",
                        help="test set(s): all,quick,unit,functional,standard,short,ubertest,cuda_memory,neuron_memory (default quick)")
    parser.add_argument("-v", dest="verbose", action="count", default=0,
                        help="verbosity level"
                             "-v: print extra info for failed test(s)"
                             "-vv: print extra info of failed/skipped test(s)"
                             "-vvv: print extra info of failed/skipped/passed test(s)")
    parser.add_argument("--expression", type=str,
                        help="only run tests which match the given substring expression.")
    parser.add_argument("--html", type=str, help="path to generated html report")
    parser.add_argument("--junit-xml", type=str, help="path to generated junit xml report")
    parser.add_argument("--junit-logging", choices=['no', 'log', 'system-out', 'system-err', 'out-err', 'all'], type=str,
                        help="Write captured log messages to JUnit report")
    parser.add_argument("--nworkers", type=int, default=8, help="Number of parallel test workers. Defaut is 8.")

    add_common_arguments(parser, shared_options)

    fabtests_args = parser.parse_args()
    if fabtests_args.provider not in ["efa", "shm"] and fabtests_args.nworkers > 1:
        print("only efa and shm provider support parallelized tests. Setting nworkers to 1 ....")
        fabtests_args.nworkers = 1

    if fabtests_args.html:
        print("html cannot be generated under parallel mode. Setting nworkers to 1 ....")
        fabtests_args.nworkers = 1

    if fabtests_args.nworkers == 1:
        exit(run(fabtests_args, shared_options, None))
    else:
        print("Running parallelable tests in parallel mode")
        parallel_status = run(fabtests_args, shared_options, "parallel")

        print("Running other tests in serial mode")
        serial_status = run(fabtests_args, shared_options, "serial")

        if fabtests_args.junit_xml:
            merged_xml = JUnitXml.fromfile(f'{fabtests_args.junit_xml}.parallel') + JUnitXml.fromfile(f'{fabtests_args.junit_xml}.serial')
            merged_xml.write(f'{fabtests_args.junit_xml}')
            os.unlink(fabtests_args.junit_xml + ".parallel")
            os.unlink(fabtests_args.junit_xml + ".serial")

        # Still return success when no tests are collected.
        if parallel_status not in [ExitCode.OK, ExitCode.NO_TESTS_COLLECTED]:
            exit(parallel_status)

        if serial_status not in [ExitCode.OK, ExitCode.NO_TESTS_COLLECTED]:
            exit(serial_status)

        exit(0)

main()



import pytest
from common import MultinodeTest


@pytest.mark.multinode
@pytest.mark.parametrize("x", ["msg", "rma"])
def test_multinode(cmdline_args, x):

    numproc = 3
    client_hostname_list = [cmdline_args.client_id, ] * (numproc - 1)
    client_base_command = "fi_multinode -x " + x + f" -n {numproc}"
    server_base_command = client_base_command
    test = MultinodeTest(cmdline_args, server_base_command, client_base_command,
                         client_hostname_list, run_client_asynchronously=True)
    test.run()

@pytest.mark.multinode
def test_multinode_coll(cmdline_args):

    numproc = 3
    client_hostname_list = [cmdline_args.client_id, ] * (numproc - 1)
    client_base_command = "fi_multinode_coll"
    server_base_command = client_base_command
    test = MultinodeTest(cmdline_args, server_base_command, client_base_command,
                         client_hostname_list, run_client_asynchronously=True)
    test.run()



from efa.efa_common import efa_run_client_server_test
from common import perf_progress_model_cli, ClientServerTest
import pytest
import copy


@pytest.mark.parametrize("iteration_type",
                         [pytest.param("short", marks=pytest.mark.short),
                          pytest.param("standard", marks=pytest.mark.standard)])
def test_rma_bw(cmdline_args, iteration_type, rma_operation_type, rma_bw_completion_semantic, rma_bw_memory_type, direct_rma_size, rma_fabric, rx_cq_data_cli):
    command = "fi_rma_bw -e rdm"
    command = command + " -o " + rma_operation_type + " " + perf_progress_model_cli + rx_cq_data_cli
    # rma_bw test with data verification takes longer to finish
    timeout = max(540, cmdline_args.timeout)
    efa_run_client_server_test(cmdline_args, command, iteration_type, rma_bw_completion_semantic,
                               rma_bw_memory_type, direct_rma_size if rma_fabric == "efa-direct" else "all",
                               timeout=timeout, fabric=rma_fabric)

@pytest.mark.parametrize("env_vars", [["FI_EFA_TX_SIZE=64"], ["FI_EFA_RX_SIZE=64"], ["FI_EFA_TX_SIZE=64", "FI_EFA_RX_SIZE=64"]])
def test_rma_bw_small_tx_rx(cmdline_args, rma_operation_type, rma_bw_completion_semantic, rma_bw_memory_type, env_vars, direct_rma_size, rma_fabric):
    cmdline_args_copy = copy.copy(cmdline_args)
    for env_var in env_vars:
        cmdline_args_copy.append_environ(env_var)
    # Use a window size larger than tx/rx size
    command = "fi_rma_bw -e rdm -W 128"
    command = command + " -o " + rma_operation_type + " " + perf_progress_model_cli
    # rma_bw test with data verification takes longer to finish
    timeout = max(540, cmdline_args_copy.timeout)
    efa_run_client_server_test(cmdline_args_copy, command, "short", rma_bw_completion_semantic,
                               rma_bw_memory_type, direct_rma_size if rma_fabric == "efa-direct" else "all",
                               timeout=timeout, fabric=rma_fabric)

@pytest.mark.functional
def test_rma_bw_range(cmdline_args, rma_operation_type, rma_bw_completion_semantic, message_size, direct_rma_size, rma_bw_memory_type, rma_fabric):
    command = "fi_rma_bw -e rdm"
    command = command + " -o " + rma_operation_type
    # rma_bw test with data verification takes longer to finish
    timeout = max(1080, cmdline_args.timeout)
    efa_run_client_server_test(cmdline_args, command, "short", rma_bw_completion_semantic,
                               rma_bw_memory_type, direct_rma_size if rma_fabric == "efa-direct" else message_size,
                               timeout=timeout, fabric=rma_fabric)


@pytest.mark.functional
def test_rma_bw_range_no_inject(cmdline_args, rma_operation_type, rma_bw_completion_semantic, inject_message_size, rma_fabric):
    if rma_fabric == "efa-direct":
        pytest.skip("Duplicate test. efa-direct has inject size = 0")
    command = "fi_rma_bw -e rdm -j 0"
    command = command + " -o " + rma_operation_type
    # rma_bw test with data verification takes longer to finish
    timeout = max(540, cmdline_args.timeout)
    efa_run_client_server_test(cmdline_args, command, "short", rma_bw_completion_semantic,
                                "host_to_host", inject_message_size, timeout=timeout, fabric=rma_fabric)


# This test is run in serial mode because it takes a lot of memory
@pytest.mark.serial
@pytest.mark.functional
@pytest.mark.parametrize("operation_type", ["read", "write", "writedata"])
def test_rma_bw_1G(cmdline_args, operation_type, rma_bw_completion_semantic, rma_fabric):
    # Default window size is 64 resulting in 128GB being registered, which
    # exceeds max number of registered host pages.
    # Use a single iteration without warmup or data verification to keep
    # CI time reasonable. Data integrity for large RMA operations is
    # covered by test_rma_bw_large below.
    timeout = max(540, cmdline_args.timeout)
    command = "fi_rma_bw -e rdm -W 1 -I 1 -w 0"
    command = command + " -o " + operation_type
    test = ClientServerTest(cmdline_args, command, iteration_type=None,
                            completion_semantic=rma_bw_completion_semantic,
                            datacheck_type="wout_datacheck",
                            message_size=1073741824,
                            memory_type="host_to_host",
                            timeout=timeout,
                            fabric=rma_fabric)
    test.run()


@pytest.mark.serial
@pytest.mark.functional
@pytest.mark.parametrize("operation_type", ["read", "write", "writedata"])
def test_rma_bw_large(cmdline_args, operation_type, rma_bw_completion_semantic, rma_fabric):
    # Verify data integrity for large RMA operations using 64MB messages.
    # This covers the same large-message code paths as 1G but completes
    # fast enough for ASAN builds.
    timeout = max(540, cmdline_args.timeout)
    command = "fi_rma_bw -e rdm -W 1"
    command = command + " -o " + operation_type
    efa_run_client_server_test(cmdline_args, command, 2,
                               completion_semantic=rma_bw_completion_semantic, message_size=67108864,
                               memory_type="host_to_host", warmup_iteration_type=0, timeout=timeout, fabric=rma_fabric)

@pytest.mark.functional
@pytest.mark.parametrize("operation_type", ["writedata", "write"])
@pytest.mark.parametrize("iteration_type",
                         ["5", # smaller than max batch wqe cnt (16)
                          "48", # larger than max batch wqe cnt
                          "128"]) # larger than window size (64)
def test_rma_bw_use_fi_more(cmdline_args, operation_type, iteration_type, rma_bw_completion_semantic, inject_message_size, direct_rma_size, rma_fabric):
    command = "fi_rma_bw -e rdm -j 0 --use-fi-more"
    command = command + " -o " + operation_type
    # rma_bw test with data verification takes longer to finish
    timeout = max(540, cmdline_args.timeout)
    efa_run_client_server_test(cmdline_args, command, iteration_type, rma_bw_completion_semantic,
                               "host_to_host", direct_rma_size if rma_fabric == "efa-direct" else inject_message_size,
                               timeout=timeout, fabric=rma_fabric)


@pytest.mark.functional
@pytest.mark.parametrize("comp_method", ["sread", "fd"])
def test_rma_bw_sread(cmdline_args, rma_operation_type, rma_bw_completion_semantic,
                      direct_rma_size, rma_bw_memory_type, support_sread, comp_method,
                      rma_fabric):
    if not support_sread:
        pytest.skip("sread not supported by efa device.")
    additional_env = ''
    if rma_fabric == "efa" and comp_method == "fd":
        if cmdline_args.server_id == cmdline_args.client_id:
            pytest.skip("FI_WAIT_FD not supported for EFA protocol with SHM enabled")
        additional_env = "FI_EFA_ENABLE_SHM_TRANSFER=0"
    command = f"fi_rma_bw -e rdm -c {comp_method}"
    command = command + " -o " + rma_operation_type
    # rma_bw test with data verification takes longer to finish
    timeout = max(1080, cmdline_args.timeout)
    efa_run_client_server_test(cmdline_args, command, "short", rma_bw_completion_semantic,
                               rma_bw_memory_type, direct_rma_size if rma_fabric == "efa-direct" else "all",
                               timeout=timeout, fabric=rma_fabric, additional_env=additional_env)



import pytest
from common import ClientServerTest

@pytest.fixture(params=[65536, 131072, 1048576])
def remote_exit_early_message_size(request):
    # 64K use medium
    # 128K use long CTS
    # 1M use runtread or longread if rdma read is available
    return request.param

@pytest.mark.functional
def test_remote_exit_early_post_send(cmdline_args, remote_exit_early_message_size):
    test = ClientServerTest(cmdline_args, "fi_efa_rdm_remote_exit_early",
                            message_size=remote_exit_early_message_size)
    test.run()

@pytest.mark.functional
def test_remote_exit_early_post_tagged(cmdline_args, remote_exit_early_message_size):
    test = ClientServerTest(cmdline_args, "fi_efa_rdm_remote_exit_early -o tagged",
                            message_size=remote_exit_early_message_size)
    test.run()

@pytest.mark.functional
def test_remote_exit_early_post_writedata(cmdline_args, remote_exit_early_message_size):
    test = ClientServerTest(cmdline_args, "fi_efa_rdm_remote_exit_early -o writedata",
                            message_size=remote_exit_early_message_size)
    test.run()

@pytest.mark.functional
def test_remote_exit_early_post_rx(cmdline_args, remote_exit_early_message_size):
    test = ClientServerTest(cmdline_args,
                            "fi_efa_rdm_remote_exit_early --post-rx",
                            message_size=remote_exit_early_message_size)
    test.run()



import pytest
from efa.efa_common import has_rdma

# this test must be run in serial mode because it will open the maximal number
# of cq that efa device can support
@pytest.mark.serial
@pytest.mark.unit
def test_cq(cmdline_args, fabric):
    from common import UnitTest
    test = UnitTest(cmdline_args, f"fi_cq_test -f {fabric}")
    test.run()

@pytest.mark.functional
@pytest.mark.parametrize("operation_type", ["senddata", "writedata"])
def test_cq_data(cmdline_args, operation_type, fabric):
    from common import ClientServerTest
    if fabric == "efa-direct" and operation_type == "writedata" and not has_rdma(cmdline_args, operation_type):
        pytest.skip("FI_RMA is not supported. Skip writedata test on efa-direct.")
    test = ClientServerTest(cmdline_args, f"fi_cq_data -e rdm -o " + operation_type, fabric=fabric)
    test.run()



import pytest
from shm.shm_common import shm_run_client_server_test
from common import perf_progress_model_cli


@pytest.mark.parametrize("operation_type", ["read", "writedata", "write"])
@pytest.mark.parametrize("iteration_type",
                         [pytest.param("short", marks=pytest.mark.short),
                          pytest.param("standard", marks=pytest.mark.standard)])
def test_rma_bw(cmdline_args, iteration_type, operation_type, completion_semantic, memory_type):
    command = "fi_rma_bw -e rdm"
    command = command + " -o " + operation_type + " " + perf_progress_model_cli
    # rma_bw test with data verification takes longer to finish
    timeout = max(540, cmdline_args.timeout)
    shm_run_client_server_test(cmdline_args, command, iteration_type, completion_semantic, memory_type, "all", timeout=timeout)

@pytest.mark.functional
@pytest.mark.parametrize("operation_type", ["read", "writedata", "write"])
def test_rma_bw_range(cmdline_args, operation_type, completion_semantic, message_size, memory_type):
    command = "fi_rma_bw -e rdm"
    command = command + " -o " + operation_type
    # rma_bw test with data verification takes longer to finish
    timeout = max(540, cmdline_args.timeout)
    shm_run_client_server_test(cmdline_args, command, "short", completion_semantic, memory_type, message_size, timeout=timeout)


from efa.efa_common import efa_run_client_server_test
from common import perf_progress_model_cli
import pytest


@pytest.fixture(params=["r:4048,4,4148",
                        "r:8000,4,9000",
                        "r:17000,4,18000"])
def rma_pingpong_message_size(request):
    return request.param


@pytest.mark.parametrize("operation_type", ["writedata"])
@pytest.mark.parametrize("iteration_type",
                         [pytest.param("short", marks=pytest.mark.short),
                          pytest.param("standard", marks=pytest.mark.standard)])
def test_rma_pingpong(cmdline_args, iteration_type, operation_type, rma_bw_completion_semantic, memory_type_bi_dir, direct_rma_size, rma_fabric):
    command = "fi_rma_pingpong -e rdm"
    command = command + " -o " + operation_type + " " + perf_progress_model_cli
    efa_run_client_server_test(cmdline_args, command, iteration_type, rma_bw_completion_semantic,
                                memory_type_bi_dir, direct_rma_size if rma_fabric == "efa-direct" else "all", fabric=rma_fabric)


@pytest.mark.functional
@pytest.mark.parametrize("operation_type", ["writedata"])
def test_rma_pingpong_range(cmdline_args, operation_type, rma_bw_completion_semantic, rma_pingpong_message_size,
                            direct_rma_size, memory_type_bi_dir, rma_fabric):
    command = "fi_rma_pingpong -e rdm"
    command = command + " -o " + operation_type
    efa_run_client_server_test(cmdline_args, command, "short", rma_bw_completion_semantic,
                               memory_type_bi_dir, direct_rma_size if rma_fabric == "efa-direct" else rma_pingpong_message_size, fabric=rma_fabric)


@pytest.mark.functional
@pytest.mark.parametrize("operation_type", ["writedata"])
def test_rma_pingpong_range_no_inject(cmdline_args, operation_type, rma_bw_completion_semantic, rma_pingpong_message_size, memory_type_bi_dir, rma_fabric):
    if rma_fabric == "efa-direct":
        pytest.skip("Duplicate test. efa-direct has inject size = 0")
    command = "fi_rma_pingpong -e rdm -j 0"
    command = command + " -o " + operation_type
    efa_run_client_server_test(cmdline_args, command, "short", rma_bw_completion_semantic,
                                memory_type_bi_dir, rma_pingpong_message_size, fabric=rma_fabric)



import pytest
import subprocess
import ctypes


def is_ibv_fork_support_needed():
    '''
    determine whether ibv(rdma-core)'s fork support is needed
    '''
    IBV_FORK_UNNEEDED = 2    # this value was taken from /usr/include/infiniband/verbs.h
    libibverbs = ctypes.cdll.LoadLibrary("libibverbs.so")
    # ibv (rdma-core)'s fork support is needed when
    # kernel space fork support is not available.
    # Newer version of rdma-core provided an API
    #      ibv_is_fork_initialized()
    # to query whether rdma-core's fork support is needed.
    # Older verison of rdma-core does not support this
    # API, so we always consider the fork support to
    # be needed.
    if hasattr(libibverbs, "ibv_is_fork_initialized"):
        return libibverbs.ibv_is_fork_initialized() != IBV_FORK_UNNEEDED
    else:
        return True


@pytest.mark.unit
def test_fork_huge_page_both_set(cmdline_args):
    """
    verify that when FI_EFA_FORK_SAFE and FI_EFA_USE_HUGE_PAGE was both set,
    application will abort
    """
    command = cmdline_args.populate_command("fi_mr_test", "host", additional_environment="FI_EFA_FORK_SAFE=1 FI_EFA_USE_HUGE_PAGE=1")
    process = subprocess.run(command, shell=True, stdout=subprocess.PIPE, stderr=subprocess.PIPE)
    if is_ibv_fork_support_needed():
        assert process.returncode != 0
        err_msg = process.stderr.decode("utf-8")
        assert "The usage of huge page is incompatible with rdma-core's fork support" in err_msg
        assert "Your application will now abort" in err_msg
    else:
        assert process.returncode == 0


#!/usr/bin/env python3

import argparse, builtins, os, sys, yaml, socket

def parse_args():
	parser = argparse.ArgumentParser(description="libfabric multinode test with slurm")
	parser.add_argument('--dry-run', action='store_true', help='Perform a dry run without making any changes.')
	parser.add_argument("--ci", type=str, help="Commands to prepend to test call. Only used with the internal launcher option", default="")
	parser.add_argument("-x", "--capability", type=str, help="libfabric capability", default="msg")
	parser.add_argument("-i", "--iterations", type=int , help="Number of iterations", default=1)
	parser.add_argument("-l", "--launcher", type=str, choices=['internal', 'srun', 'mpirun'], help="launcher to use for running job. If nothing is specified, test manages processes internally. Available options: internal, srun and mpirun", default="internal")

	required = parser.add_argument_group("Required arguments")
	required.add_argument("-p", "--provider", type=str, help="libfabric provider")
	required.add_argument("-np", "--num-procs", type=int, help="Map process by node, l3cache, etc")
	required.add_argument("-c", "--config", type=str, help="Test configuration")

	srun_required = parser.add_argument_group("Required if using srun")
	srun_required.add_argument("-t", "--procs-per-node", type=int,
							help="Number of procs per node", default=-1)

	args = parser.parse_args()
	return args, parser

def parse_config(config):
	with open(config, "r") as f:
		yconf = yaml.safe_load(f)
	return yconf

def mpi_env(config):
	env = config['environment']
	result = []
	for k in env.keys():
		result.append(f"-x {k}")
	return " ".join(result)

def set_env(config):
	env = config['environment']
	for k, v in env.items():
		os.environ[k] = str(v)

def mpi_mca_params(config):
	try:
		mca = config['mca']
		result = []
		for k, v in env.items():
			result.append(f"--mca {k} {v}")
		return " ".join(result)
	except:
		return ""

def mpi_bind_to(config):
	try:
		return f"--bind-to {config['bind-to']}"
	except:
		return "--bind-to core"

def mpi_map_by(config):
	try:
		return f"--map-by ppr:{config['map-by-count']}:{config['map-by']}"
	except:
		return "--map-by ppr:1:l3"

def execute_cmd(cmd, dry_run):
	script_dir = os.path.dirname(os.path.abspath(__file__))
	sys.path.append(script_dir)
	from command import Command
	cmd = Command(cmd, fake=dry_run)
	rc, out = cmd.exec_cmd()
	return rc, out

def split_on_commas(expr): 
	l = []
	c = o = 0 
	s = expr
	stop = False
	while not stop and c != -1 and o != -1:
		o = s.find('[')
		b = s.find(']')
		c = s.find(',')
		while c > o and c < b:
			c = s.find(',', c+1)
		if len(l):
			l.pop()
		if c < o or c > b:
			l += [s[:s.find(',', c)], s[s.find(',', c)+1:]]
			s = l[-1]
		else:
			l += s.split(',')
			stop = True
	for i in range(0, len(l)):
		l[i] = l[i].strip()
	return l

def expand_host_list_sub(expr):
	host_list = []

	# first phase split on the commas first
	open_br = expr.find('[')
	close_br = expr.find(']', open_br)
	if open_br == -1 and close_br == -1:
		return expr.split(',')

	if open_br == -1 or close_br == -1:
		return []

	rangestr = expr[open_br+1 : close_br]

	node = expr[:open_br]

	ranges = rangestr.split(',')

	for r in ranges:
		cur = r.split('-')
		if len(cur) == 2:
			pre = "{:0%dd}" % len(cur[0])
			for idx in range(int(cur[0]), int(cur[1])+1):
				host_list.append(f'{node}{pre.format(idx)}')
		elif len(cur) == 1:
			pre = "{:0%dd}" % len(cur[0])
			host_list.append(f'{node}{pre.format(int(cur[0]))}')

	return host_list

def expand_host_list(expr):
	l = split_on_commas(expr)
	host_list = []
	for e in l:
		host_list += expand_host_list_sub(e)
	return host_list

supported_pm = ['pmi', 'pmi2', 'pmix']

def is_srun_pm_supported():
	rc, out = execute_cmd('srun --mpi=list', False)
	if rc:
		return False
	input_list = out.split('\n')
	cleaned_list = [entry.strip() for entry in input_list if entry.strip()]
	for e in supported_pm:
		if e in cleaned_list[1:]:
			return True
	return False

if __name__ == '__main__':

	# list of providers which do not address specification
	no_addr_prov = ['cxi']

	args, parser = parse_args()

	if not args.config:
		print("**A configuration file is required")
		print(parser.format_help())
		exit(-1)

	mnode = parse_config(args.config)['multinode']
	set_env(mnode)

	if args.launcher == 'srun':
		if not is_srun_pm_supported():
			print(f"**Supported process managers are: {','.join(supported_pm)}")
			print(parser.format_help())
			exit(-1)

	# The script assumes it's already running in a SLURM allocation. It can
	# then srun fi_multinode
	#
	if "pattern" not in mnode:
		print("Test pattern must be defined in the YAML configuration file")
		exit()

	if args.provider in no_addr_prov:
		cmd = f"fi_multinode -n {args.num_procs} -s {socket.gethostname()} " \
			f"-p {args.provider} -x {args.capability} -z {mnode['pattern']} " \
			f"-I {args.iterations} -u {args.launcher.lower()} -T"
	else:
		cmd = f"fi_multinode -n {args.num_procs} -s {socket.gethostname()} " \
			f"-p {args.provider} -x {args.capability} -z '{mnode['pattern']}' " \
			f"-I {args.iterations} -u {args.launcher.lower()} -T"

	if args.launcher.lower() == 'mpirun':
		mpi = f"mpirun {mpi_env(mnode)} {mpi_mca_params(mnode)} {mpi_bind_to(mnode)} "\
			  f"{mpi_map_by(mnode)} -np {args.num_procs} {cmd}"
	elif args.launcher.lower() == 'srun':
		if args.procs_per_node == -1 or args.num_procs == -1:
			print("**Need to specify --procs-per-node and --num-procs")
			print(parser.format_help())
			exit()
		mpi = f"srun --ntasks-per-node {args.procs_per_node} --ntasks {args.num_procs} "\
			  f"{cmd}"
	elif args.launcher.lower() == 'internal':
		if args.procs_per_node == -1:
			print("**Need to specify --procs-per-node")
			print(parser.format_help())
			exit()
		hl = ",".join(expand_host_list(os.environ['SLURM_NODELIST']))
		mpi = f"runmultinode.sh -h {hl} -n {args.procs_per_node} -p {args.provider} " \
			  f"-x {args.capability} -I {args.iterations} -z {mnode['pattern']}"
		if args.ci:
			mpi += f" --ci '{args.ci}'"
	else:
		print("**Unsupported launcher")
		print(parser.format_help())
		exit()

	rc, out = execute_cmd(mpi, args.dry_run)

	print(f"Command completed with {rc}\n{out}")



import copy
import pytest
import subprocess
from efa.efa_common import efa_retrieve_hw_counter_value, get_efa_device_names
from common import ClientServerTest

# This test must be run in serial mode because it checks the hw counter
@pytest.mark.serial
@pytest.mark.functional
@pytest.mark.parametrize("selection_approach", ["domain name", "environment"])
def test_efa_device_selection(cmdline_args, fabric, selection_approach):

    if cmdline_args.server_id == cmdline_args.client_id:
        pytest.skip("EFA device selection test requires 2 nodes")
        return

    server_device_names = get_efa_device_names(cmdline_args.server_id)
    client_device_names = get_efa_device_names(cmdline_args.client_id)

    server_num_devices = len(server_device_names)
    client_num_devices = len(client_device_names)
    assert server_num_devices > 0, "No EFA devices found on server"
    assert client_num_devices > 0, "No EFA devices found on client"

    for i in range(max(server_num_devices, client_num_devices)):
        server_device_idx = i % server_num_devices
        client_device_idx = i % client_num_devices

        server_device_name = server_device_names[server_device_idx]
        client_device_name = client_device_names[client_device_idx]

        for suffix in ["rdm", "dgrm"]:
            if fabric == "efa-direct" and suffix == "dgrm":
                continue
            server_tx_bytes_before_test = efa_retrieve_hw_counter_value(cmdline_args.server_id, "tx_bytes", server_device_name)
            client_tx_bytes_before_test = efa_retrieve_hw_counter_value(cmdline_args.client_id, "tx_bytes", client_device_name)

            prefix_type = "wout_prefix"
            strict_fabtests_mode = True
            if suffix == "rdm":
                command = "fi_rdm_pingpong"
            else:
                prefix_type = "with_prefix"  # efa provider requires prefix mode for dgram provider, hence "-k"
                strict_fabtests_mode = False  # # dgram is unreliable
                command = "fi_dgram_pingpong"

            server_domain_name = server_device_name + "-" + suffix
            client_domain_name = client_device_name + "-" + suffix

            cmdline_args_copy = copy.copy(cmdline_args)
            if selection_approach == "domain name":
                cmdline_args_copy.additional_server_arguments = "-d " + server_domain_name
                cmdline_args_copy.additional_client_arguments = "-d " + client_domain_name
            else:
                assert selection_approach == "environment"
                cmdline_args_copy.append_server_environ(f"FI_EFA_IFACE={server_device_name}")
                cmdline_args_copy.append_client_environ(f"FI_EFA_IFACE={client_device_name}")
            cmdline_args_copy.strict_fabtests_mode = strict_fabtests_mode

            test = ClientServerTest(cmdline_args_copy, command, message_size="1000", prefix_type=prefix_type, timeout=300, fabric=fabric, iteration_type="short")
            test.run()

            server_tx_bytes_after_test = efa_retrieve_hw_counter_value(cmdline_args.server_id, "tx_bytes", server_device_name)
            client_tx_bytes_after_test = efa_retrieve_hw_counter_value(cmdline_args.client_id, "tx_bytes", client_device_name)

            # Verify EFA traffic
            assert server_tx_bytes_before_test < server_tx_bytes_after_test
            assert client_tx_bytes_before_test < client_tx_bytes_after_test

# Verify that fi_getinfo does not return any info objects when FI_EFA_IFACE is set to an invalid value
@pytest.mark.functional
def test_efa_device_selection_negative(cmdline_args, fabric):
    invalid_iface = "r"

    command = cmdline_args.populate_command(f"fi_efa_info_test -f {fabric}", "host", additional_environment=f"FI_EFA_IFACE={invalid_iface}")
    proc = subprocess.run(command, shell=True,
                          stdout=subprocess.PIPE, stderr=subprocess.PIPE,
                          encoding="utf-8", timeout=60)
    assert proc.returncode == 61

# Verify that fi_getinfo returns all NICs when FI_EFA_IFACE is set to all
@pytest.mark.functional
def test_efa_device_selection_all(cmdline_args, fabric):
    devices = get_efa_device_names(cmdline_args.server_id)
    assert len(devices) > 0, "No EFA devices found on server"

    command = cmdline_args.populate_command(f"fi_efa_info_test -f {fabric}", "host", additional_environment="FI_EFA_IFACE=all")
    proc = subprocess.run(command, shell=True,
                          stdout=subprocess.PIPE, stderr=subprocess.PIPE,
                          encoding="utf-8", timeout=60)
    assert proc.returncode == 0

    num_domains = 0
    info_out = proc.stdout.strip().split("\n")
    for info in info_out:
        if "domain" in info:
            num_domains += 1

    assert num_domains == len(devices)

# Verify that fi_getinfo returns two NICs when FI_EFA_IFACE is set to two NICs separated by a comma
@pytest.mark.functional
def test_efa_device_selection_comma(cmdline_args, fabric):
    devices = get_efa_device_names(cmdline_args.server_id)
    assert len(devices) > 0, "No EFA devices found on server"

    if len(devices) > 1:
        iface_str = f"{devices[0]},{devices[1]}"
    else:
        iface_str = f"{devices[0]},{devices[0]}"

    command = cmdline_args.populate_command(f"fi_efa_info_test -f {fabric}", "host", additional_environment=f"FI_EFA_IFACE={iface_str}")
    proc = subprocess.run(command, shell=True,
                          stdout=subprocess.PIPE, stderr=subprocess.PIPE,
                          encoding="utf-8", timeout=60)
    assert proc.returncode == 0

    num_domains = 0
    info_out = proc.stdout.strip().split("\n")
    for info in info_out:
        if "domain" in info:
            num_domains += 1

    if len(devices) > 1:
        assert num_domains == 2
    else:
        assert num_domains == 1



import copy

import pytest
from common import UnitTest, has_cuda, has_neuron


@pytest.mark.unit
def test_mr_host(cmdline_args):
    test = UnitTest(cmdline_args, "fi_mr_test")
    test.run()


@pytest.mark.unit
@pytest.mark.short
def test_mr_hmem(cmdline_args, hmem_type, fabric):
    if hmem_type == "cuda" and not has_cuda(cmdline_args.server_id):
        pytest.skip("no cuda device")
    if hmem_type == "neuron" and not has_neuron(cmdline_args.server_id):
        pytest.skip("no neuron device")

    cmdline_args_copy = copy.copy(cmdline_args)

    test_command = f"fi_mr_test -D {hmem_type} -f {fabric}"

    if cmdline_args.do_dmabuf_reg_for_hmem:
        test_command += " -R"

    test = UnitTest(
        cmdline_args_copy,
        test_command,
        failing_warn_msgs=["Unable to add MR to map"],
    )
    test.run()


@pytest.mark.unit
@pytest.mark.short
def test_efa_mr_hmem(cmdline_args, hmem_type, fabric):
    if hmem_type != "neuron":
        pytest.skip("test only applies to neuron")
    if not has_neuron(cmdline_args.server_id):
        pytest.skip("no neuron device")
    if not cmdline_args.do_dmabuf_reg_for_hmem:
        pytest.skip("This test tests neuron get_dmabuf_fd_vX and needs dmabuf to be enabled and working to test these apis")

    cmdline_args_copy = copy.copy(cmdline_args)

    test_command = f"fi_efa_mr_test -D neuron -f {fabric}"

    if cmdline_args.do_dmabuf_reg_for_hmem:
        test_command += " -R"

    test = UnitTest(cmdline_args_copy, test_command)
    test.run()



import pytest
from common import ClientServerTest

@pytest.mark.functional
@pytest.mark.parametrize("shared_cq", [True, False])
@pytest.mark.parametrize("shared_av", [True, False])
def test_multi_ep(cmdline_args, shared_cq, shared_av, rma_fabric):
    # This test requests FI_RMA
    cmd = "fi_multi_ep -e rdm"
    if shared_cq:
        cmd += " -Q"
    if shared_av:
        cmd += " -A"
    test = ClientServerTest(cmdline_args, cmd, message_size=256, fabric=rma_fabric)
    test.run()


# TODO: expand to different meory types
@pytest.mark.unstable
@pytest.mark.parametrize("msg_size", [16, 8192, 16384, 1048576])
@pytest.mark.parametrize("shared_av", [True, False])
def test_multi_ep_mt(cmdline_args, msg_size, shared_av):
    cmd = f"fi_efa_multi_ep_mt -e rdm -c 20 -I 100"

    if shared_av:
        cmd += " -A"

    test = ClientServerTest(cmdline_args, cmd, message_size=msg_size, fabric="efa", timeout=30, additional_env="FI_EFA_ENABLE_SHM_TRANSFER=0")
    test.run()



import pytest
from default.test_rdm import test_rdm, \
    test_rdm_bw_functional, test_rdm_atomic
from common import perf_progress_model_cli
from shm.shm_common import shm_run_client_server_test

@pytest.mark.parametrize("iteration_type",
                         [pytest.param("short", marks=pytest.mark.short),
                          pytest.param("standard", marks=pytest.mark.standard)])
def test_rdm_pingpong(cmdline_args, iteration_type, completion_semantic, memory_type, completion_type):
    command = "fi_rdm_pingpong" + " " + perf_progress_model_cli
    shm_run_client_server_test(cmdline_args, command, iteration_type,
                               completion_semantic, memory_type, completion_type=completion_type)


@pytest.mark.parametrize("iteration_type",
                         [pytest.param("short", marks=pytest.mark.short),
                          pytest.param("standard", marks=pytest.mark.standard)])
def test_rdm_tagged_pingpong(cmdline_args, iteration_type, completion_semantic, memory_type, completion_type):
    command = "fi_rdm_tagged_pingpong" + " " + perf_progress_model_cli
    shm_run_client_server_test(cmdline_args, command, iteration_type,
                               completion_semantic, memory_type, completion_type=completion_type)


@pytest.mark.parametrize("iteration_type",
                         [pytest.param("short", marks=pytest.mark.short),
                          pytest.param("standard", marks=pytest.mark.standard)])
def test_rdm_tagged_bw(cmdline_args, iteration_type, completion_semantic, memory_type, completion_type):
    command = "fi_rdm_tagged_bw" + " " + perf_progress_model_cli
    shm_run_client_server_test(cmdline_args, command, iteration_type,
                               completion_semantic, memory_type, completion_type=completion_type)

@pytest.mark.functional
def test_rdm_tagged_peek(cmdline_args):
    from copy import copy

    from common import ClientServerTest

    test = ClientServerTest(cmdline_args, "fi_rdm_tagged_peek", timeout=1800)
    test.run()


from common import MultinodeTest
import pytest


def test_efa_shm_addr(cmdline_args, fabric):
    server_id = cmdline_args.server_id
    client_id = cmdline_args.client_id
    if client_id == server_id:
        pytest.skip("This test requires two nodes")
    # First start a client on remote host, then start
    # a client on local host, so the shm fi_addr
    # inserted for the 2nd client could be different
    # from its efa fi_addr.
    client_hostname_list = [client_id, server_id]
    client_base_command = f"fi_rdm -f {fabric}"
    server_base_command = client_base_command + " -C {}".format(len(client_hostname_list))
    test = MultinodeTest(cmdline_args, server_base_command, client_base_command,
                         client_hostname_list, run_client_asynchronously=False)
    test.run()



#!/usr/bin/env python3
# SPDX-License-Identifier: BSD-2-Clause OR GPL-2.0-only
# Copyright (c) 2021-2023 Hewlett Packard Enterprise Development LP

help = f'''
Multicast REST server simulation for distributed testing

http://host:port/
- GET produces this help as a JSON list

http://host:port/fabric/collectives/multicast
- POST generates a single multicast address and hwroot node
- GET lists all multicast addresses
- DELETE deletes all multicast addresses

http://host:port/fabric/collectives/mcastid/<id>
- DELETE deletes specified multicast address

Multicast addresses are invalid (>=8192), causing UNICAST behavior
Only addresses 8192-8199 are supported, to test exhaustion
'''
import argparse
import textwrap
import sys
import json

from argparse import ArgumentParser, HelpFormatter
from flask import Flask, request
from flask_restful import Api, Resource

# Global storage for addresses/roots
mcastroots = []
mcastaddrs = []

class RawFormatter(HelpFormatter):
    def _fill_text(self, text, width, indent):
        return "\n".join([textwrap.fill(line, width) for line in textwrap.indent(textwrap.dedent(text), indent).splitlines()])

class fabtestInfo(Resource):
    def get(self):
        return help.splitlines(), 200

def delEntry(value):
    global mcastroots
    global mcastaddrs

    try:
        idx = mcastaddrs.index(value)
        del mcastroots[idx]
        del mcastaddrs[idx]
        print("DELETE ", value)
    except:
        print("multicast", value, "not in use")
        pass

class delete8192(Resource):
    def delete(self):
        delEntry(8192)

class delete8193(Resource):
    def delete(self):
        delEntry(8193)

class delete8194(Resource):
    def delete(self):
        delEntry(8194)

class delete8195(Resource):
    def delete(self):
        delEntry(8195)

class delete8196(Resource):
    def delete(self):
        delEntry(8196)

class delete8197(Resource):
    def delete(self):
        delEntry(8197)

class delete8198(Resource):
    def delete(self):
        delEntry(8198)

class delete8199(Resource):
    def delete(self):
        delEntry(8199)

class fabtestServer(Resource):
    def get(self):
        # Lists the existing multicast addresses
        global mcastroots
        global mcastaddrs

        addrs = []
        for k,v in enumerate(mcastroots):
            addrs.append({'root':v, 'mcast':mcastaddrs[k]})
        info = {
            'ADDRLIST': addrs,
        }
        return info, 200

    def delete(self):
        # Deletes all multicast addresses
        global mcastroots
        global mcastaddrs

        mcastroots = []
        mcastaddrs = []
        return None, 200

    def post(self):
        # Creates a new multicast address
        global mcastroots
        global mcastaddrs

        print(request.json)
        required = {
            'jobID', 'macs', 'timeout',
        }
        optional = {
            'jobStepID'
        }
        info = {}
        error = []
        dupmac = []

        # Test for required fields, append error messages if missing
        for key in required:
            if key not in request.json:
                error.append("no " + key)
            else:
                info[key] = request.json[key]
        # Test macs for empty or duplicate addresses
        if not error and not request.json['macs']:
            error.append('empty macs')
        for mac in request.json['macs']:
            if mac not in dupmac:
                dupmac.append(mac)
            else:
                error.append('duplicate mac=' + str(mac))

        # Test for optional fields, provide defaults if missing
        for key in optional:
            if key not in request.json:
                info[key] = None
            else:
                info[key] = request.json[key]

        # Find a globally-unused mac address as hwRoot
        info['hwRoot'] = None
        for mac in request.json['macs']:
            if mac not in mcastroots:
                info['hwRoot'] = mac
                break
        if not info['hwRoot']:
            error.append('no hwRoot usable')

        # Find a globally unused mcast address
        info['mcastID'] = None
        for adr in range(8192, 8199):
            if adr not in mcastaddrs:
                info['mcastID'] = adr
                break
        if not info['mcastID']:
            error.append('no mcast available')

        # Report any accumulated errors
        if error:
            info = {
                'error' : ', '.join(error)
            }
            return info, 400

        # Otherwise, record and return complete record
        mcastroots.append(mac)
        mcastaddrs.append(adr)

        info['jobID'] = request.json['jobID']
        info['jobStepID'] = request.json['jobStepID']
        info['macs'] = request.json['macs']
        info['timeout'] = request.json['timeout']
        info['documentSelfLink'] = 'fabric/collectives/mcastID/' + adr

        return info, 200

def main(argv):
    parser = argparse.ArgumentParser(
        description=help, formatter_class=RawFormatter)
    parser.add_argument('--host', default=None)
    parser.add_argument('--port', default=None)
    args = parser.parse_args()

    app = Flask(__name__)
    api = Api(app);
    api.add_resource(fabtestInfo, '/')
    api.add_resource(fabtestServer, '/fabric/collectives/multicast')
    api.add_resource(delete8192, '/fabric/collectives/mcastid/8192')
    api.add_resource(delete8193, '/fabric/collectives/mcastid/8193')
    api.add_resource(delete8194, '/fabric/collectives/mcastid/8194')
    api.add_resource(delete8195, '/fabric/collectives/mcastid/8195')
    api.add_resource(delete8196, '/fabric/collectives/mcastid/8196')
    api.add_resource(delete8197, '/fabric/collectives/mcastid/8197')
    api.add_resource(delete8198, '/fabric/collectives/mcastid/8198')
    api.add_resource(delete8199, '/fabric/collectives/mcastid/8199')
    app.run(debug=True, host=args.host, port=args.port)

if __name__ == "__main__":
    main(sys.argv)



#!/usr/bin/env python3

"""
Executes criterion tests using parameters defined in an 
input yaml file and outputs results to a file in TAP format
"""

import pexpect
import sys
import yaml
import re

from enum import Enum
from argparse import ArgumentParser
from contextlib import ExitStack


class Result(Enum):
    """
    Enum for test results
    """
    PASSED = 1
    FAILED = 2
    SKIPPED = 3


class Node:
    """
    Class for managing a node's SSH connection
    """
    def __init__(self, node_name):
        """
        Args:
            node_name: the name of the node
        """
        self.name = node_name
        self.ssh = open_ssh(self.name, prompt=PROMPT)
        self.ssh.logfile = sys.stdout


class TestSet:
    """
    Set of tests that use common runtime parameters and CSR settings
    """
    def __init__(self, desc, test_filter=None, runtime_params=None, csr_list=None):
        """
        Args:
            desc: description of the test set
            test_filter: tests to run
            runtime_params: runtime parameters to use with this test set
            csr_list: csrs to set prior to running the tests
        """
        self.description = desc
        self.filter = test_filter
        self.runtime_params = runtime_params
        self.csr_list = csr_list
        self.csr_list_initial_vals = []
        self.test_list = []

        # generate the list of tests
        self.generate_testlist()

    def set_csrs_for_test_set(self):
        """
        capture original csr values and set csrs to new values

        """
        if self.csr_list is not None:
            send(node, "cd {}".format(PYCXI_DIR))

            for c in self.csr_list:
                csr = c[0]
                field = c[1]
                new_value = c[2]

                # capture initial values
                orig_val = get_csr_value(csr, field)
                self.csr_list_initial_vals.append([csr, field, orig_val])

                # set new values
                set_csr_value(csr, field, new_value)

    def restore_csrs(self):
        """
        Restore csrs to their original values

        """
        if self.csr_list is not None:
            print("Restoring CSRs...")
            send(node, "cd {}".format(PYCXI_DIR))

            for c in self.csr_list_initial_vals:
                field = c[1]
                if isinstance(c[2], dict):
                    # csr is an array, so set each member of the array
                    for csr, orig_value in c[2].items():
                        set_csr_value(csr, field, orig_value)
                else:
                    csr = c[0]
                    orig_value = c[2]
                    set_csr_value(csr, field, orig_value)

    def generate_testlist(self):
        """
        generate a list of tests to run based on the provided filter

        """
        send(node, "cd {}".format(TEST_DIR))

        # create criterion test list
        send(node, './cxitest -l > testlist 2>&1 && sleep 1 && echo "DONE"', resp_1="DONE")
        f_name = "{}/testlist".format(TEST_DIR)
        with open(f_name) as file:
            all_lines = [line.rstrip() for line in file.readlines()]

        # create regex instance for filter (if needed)
        regex_filter = None
        if self.filter is not None:
            regex_filter = re.compile(self.filter, re.IGNORECASE)

        # parse testlist and create test objects for this test set
        area = None
        for line in all_lines:
            if ':' in line:
                area = line.split()[0].replace(":", "")
            else:
                tst_name = line.split()[1]

                # indicates "disabled" flag was set in Criterion test
                skip_test = "skipped" in line

                # create test objects for this test set based on the filter (if provided)
                if self.filter is None or (
                        regex_filter is not None and regex_filter.match("{}/{}".format(area, tst_name))):
                    tst = Test(area, tst_name, self.description, self.runtime_params, skip=skip_test)
                    self.test_list.append(tst)

    def execute_tests(self):
        """
        Executes the tests in the test set and capture the output
        """
        with ExitStack() as cleanup:

            # restore CSRs on exit
            cleanup.callback(self.restore_csrs)

            # set CSRs for test set
            self.set_csrs_for_test_set()

            send(node, "cd {}".format(TEST_DIR))

            # execute tests in test list
            for te in self.test_list:
                sys.stdout.flush()
                cmd = '{} > tmp_result 2>&1 && echo "DONE"'.format(te.test_cmd)

                # execute test
                send(node, cmd, resp_1="DONE", timeout=60)

                results_index = 0
                enable_logging = False

                # process raw results file
                with open("{}/tmp_result".format(TEST_DIR)) as file:
                    all_lines = [line.strip() for line in file.readlines()]

                # capture all output related to this test
                for ln in all_lines:
                    line = ansi_escape.sub('', ln).rstrip()
                    test_str = " {}::{}".format(te.test_area, te.test_name)
                    if test_str in line:
                        if "RUN" in line and line.endswith(test_str):
                            # start capturing output for this test
                            enable_logging = True

                            # create a TestResult instance for this test
                            te.results.append(TestResult(results_index))

                            # if CSRs were modified, include that in the log:
                            if self.csr_list is not None:
                                te.results[results_index].log.append("Modified CSRs: {}".format(self.csr_list))

                            # log the actual Criterion test command
                            te.results[results_index].log.append("Test cmd: {}".format(te.test_cmd))

                            # log the "RUN" output
                            te.results[results_index].log.append(line)
                        elif "{}:".format(test_str) in line:

                            # set the test result
                            te.results[results_index].result = get_result(line)

                            # capture the entire result line
                            te.results[results_index].log.append(line)

                            # the test is finished, so stop capturing output for this test
                            enable_logging = False

                            # increment index (multiple results for Test instance indicates a parameterized tests)
                            results_index += 1

                    elif enable_logging:
                        # test is in process, so capture all console output that occurs
                        te.results[results_index].log.append(line)

                # display all logged output belonging to this particular test
                for res in te.results:
                    print("\n-------------------------------------------------------")
                    for s in res.log:
                        print(s)
                    print("-------------------------------------------------------\n")
    

class Test:
    """
    An individual test, which may contain multiple TestResult objects if the test is parameterized
    """
    def __init__(self, test_area, test_name, desc, t_params=None, skip=False):
        """
        Args:
            test_area: the test area
            test_name: the test name
            desc: description of the test
            t_params: runtime parameters for this test
            skip: flag to indicate if the test should be skipped
        """
        self.test_area = test_area
        self.test_name = test_name
        self.desc = desc
        self.skip = skip
        self.results = []

        # create the runtime parameters string for this test
        param_str = ""
        if t_params is not None:
            for pa, v in t_params.items():
                param_str += "{}={} ".format(pa, v)

        self.test_params = param_str

        # create the test cmd
        self.test_cmd = \
            '{} ./cxitest --filter="{}/{}" --verbose=1 -j1 --ascii'.format(param_str, test_area, test_name)

        # create TestResult for skipped test
        if self.skip:
            st = TestResult()
            st.result = Result.SKIPPED
            self.results.append(st)

    def create_tap_results(self):
        """
        Parse results log and create TAP results for this test
        """
        for res in self.results:
            # get test number for this test
            test_num = get_current_test_count_and_inc()

            # determine TAP result based on test result
            tap_result = "ok {}".format(test_num) if res.result != Result.FAILED else "not ok {}".format(test_num)

            # construct the TAP test name
            t_name = "{}::{}".format(self.test_area, self.test_name)

            # if we have a parameterized test, append index to the test name
            if len(self.results) > 1:
                t_name = "{}::{}".format(t_name, res.index)

            # append the description
            t_name = "{} - {}".format(t_name, self.desc)

            # if test was skipped, include skip comment
            if res.result == Result.SKIPPED:
                t_name = "{} # skip".format(t_name)

                # include additional comment for disabled tests
                if self.skip:
                    t_name += " Disabled flag set in criterion test "

            # append the tap result and test name to the tap report
            tap_report.append("{} {}".format(tap_result, t_name))

            # include all logged output during this test in the tap report
            for m in res.log:
                tap_report.append("# {}".format(m))


class TestResult:
    """
    Result and log for a particular test
    """
    def __init__(self, index=0):
        """

        Args:
            index: test index - used with parameterized tests
        """
        self.index = index
        self.result = Result.FAILED
        self.log = []


def get_result(the_line):
    """
    Determine the test result from the given line

    Args:
        the_line: the line to check

    Returns: the result

    """
    if "PASS" in the_line:
        return Result.PASSED
    elif "SKIP" in the_line:
        return Result.SKIPPED
    else:
        return Result.FAILED


def set_csr_value(csr, field, value):
    """
    Sets a CSR field to the given value

    Args:
        csr: the CSR
        field: the field
        value: the value

    """
    # use cxiutil to set the value
    send(node, "cd {}".format(PYCXI_DIR))
    cmd = 'cxiutil store csr {} {}={} && sleep 1 && echo "DONE"'.format(csr, field, value)
    send(node, cmd, resp_1="DONE")
    sys.stdout.flush()

    # verify the new value is set as expected
    new_val = get_csr_value(csr, field)
    if isinstance(new_val, dict):
        # we have a CSR array, so verify each member of the array
        for v in new_val.values():
            if int(v) != int(value):
                raise RuntimeError("Unable to set CSR with cmd: {}. "
                                   "Actual value of {} = {}".format(cmd, field, v))
    else:
        if int(new_val) != int(value):
            raise RuntimeError("Unable to set CSR with cmd: {}. "
                               "Actual value of {} = {}".format(cmd, field, new_val))


def get_csr_value(csr, field):
    """
    Returns the value of the CSR field. If the CSR is an array, returns a dict containing each CSR index and value

    Args:
        csr: the CSR
        field: the field

    Returns: the value, or a dict containing each CSR index and value

    """

    # use cxiutil to get the value
    send(node, "cd {}".format(PYCXI_DIR))
    sys.stdout.flush()
    send(node, 'cxiutil dump csr {} > tmp && sleep 1 && echo "DONE"'.format(csr, field), resp_1="DONE")

    with open("{}/tmp".format(PYCXI_DIR)) as file:
        all_lines = [line.rstrip() for line in file.readlines()]

    # parse the cxiutil output
    response = {}
    for line in all_lines:
        if "hex" in line:
            csr = line.split()[0]

        if field in line and "0x" in line:
            response[csr] = line.split()[2]

    # csr array, so return a dict containing each value in the csr array
    if len(response) > 1:
        return response
    # not a csr array, so just return the value
    elif len(response) == 1:
        return response[csr]
    else:
        raise RuntimeError("Unable to read CSR {} {}".format(csr, field))


def generate_tap_file():
    """
    generate the TAP results file
    """
    total_test_count = 0

    # capture the total number of tests
    for ts in test_set_list:
        for tst in ts.test_list:
            total_test_count += len(tst.results)

    # add TAP header line
    tap_header = "1..{}".format(total_test_count)
    tap_report.append(tap_header)

    # capture TAP results of each test
    for ts in test_set_list:
        for element in ts.test_list:
            element.create_tap_results()

    # create TAP file
    with open(RESULTS_FILE, 'w') as file_handler:
        for tap_line in tap_report:
            file_handler.write("{}\n".format(tap_line))
            print(tap_line)


def get_current_test_count_and_inc():
    """
    returns the current test count prior to incrementing it

    Returns: the current test count

    """
    global current_test_count
    tmp_count = current_test_count
    current_test_count += 1
    return tmp_count


def open_ssh(node_addr, prompt):
    """
    Create ssh connection to the given ip address

    Args:
        node_addr: the node name / ip address
        prompt: the prompt to expect

    Returns: SSH connection / process

    """
    s = pexpect.spawn("ssh {}".format(node_addr), encoding='utf-8')
    try:
        rc = s.expect([prompt, "Password:"], timeout=30)
        if rc == 1:
            s.sendline(PASSWORD)
            s.expect(prompt, timeout=10)
    except pexpect.TIMEOUT:
        print("Unable to ssh to {}".format(node_addr))
        raise pexpect.TIMEOUT
    return s


def send(the_node, cmd, resp_1=None, resp_2=None, expect_prompt=True, timeout=30):
    """
    send a command to the given node and verify expected response(s)

    Args:
        the_node: the node
        cmd: the command to send
        resp_1: the first expected response (if not None)
        resp_2: the second expected response (if not None)
        expect_prompt: flag to indicate if a prompt is expected
        timeout: the maximum time to wait for a response before throwing an exception
    """
    ssh_sesh = the_node.ssh
    ssh_sesh.sendline(cmd)

    if resp_1:
        ssh_sesh.expect(resp_1, timeout=timeout)

    if resp_2:
        ssh_sesh.expect(resp_2, timeout=timeout)

    if expect_prompt:
        ssh_sesh.expect(PROMPT, timeout=timeout)


if __name__ == "__main__":

    # used to filter ansi escape chars
    ansi_escape = re.compile(r'(?:\x1B[@-_]|[\x80-\x9F])[0-?]*[ -/]*[@-~]')

    p = ArgumentParser("run_criterion_tests")
    p.add_argument('-n',
                   dest="node",
                   nargs='?',
                   type=str,
                   required=True,
                   help="Name of node where test is to be run")

    p.add_argument('-y',
                   dest="yaml_file",
                   nargs='?',
                   type=str,
                   required=True,
                   help="Path to the test YAML file")

    args = p.parse_args()

    # parse the input yaml file
    try:
        with open(args.yaml_file, 'r') as stream:
            f = yaml.safe_load(stream)

    except FileNotFoundError:
        print("YAML file not found: {}".format(args.yaml_file))

    LIBFABRIC_DIR = f["env"]["libfabric_dir_on_node"]
    TEST_DIR = "{}/prov/cxi/test".format(LIBFABRIC_DIR)
    PYCXI_DIR = f["env"]["pycxi_dir_on_node"]
    PROMPT = f["env"]["node_prompt"]
    PASSWORD = f["env"]["node_password"]
    RESULTS_FILE = "{}/results.tap".format(TEST_DIR)

    # holds all TAP results
    tap_report = []

    # instantiate node object
    node = Node(args.node)

    # activate pycxi venv for cxiutil and remove old tap files
    send(node, "cd {}".format(PYCXI_DIR))
    send(node, ". .venv/bin/activate")
    send(node, "cd {}".format(TEST_DIR))
    send(node, "rm *.tap")

    # set global runtime parameters prior to running tests
    default_runtime_parameters = f["global_runtime_parameters"]
    for params in default_runtime_parameters:
        for param, val in params.items():
            send(node, "export {}={}".format(param, val))

    current_test_count = 1

    # create test sets
    test_set_list = []
    for test in f["tests"]:
        test_set_list.append(TestSet(
            desc=test["description"],
            test_filter=test["filter"],
            runtime_params=test["runtime_parameters"],
            csr_list=test["csrs"])
        )

    # execute the tests in each test set
    for test_set in test_set_list:
        test_set.execute_tests()

    # generate the tap file
    generate_tap_file()

